import dataclasses
import logging
from typing import ClassVar, List, Tuple

import numpy as np
import torch

from .annrescaler import AnnRescaler
from .. import headmeta
from ..visualizer import Caf as CafVisualizer
from ..utils import mask_valid_area

LOG = logging.getLogger(__name__)


@dataclasses.dataclass
class Caf:
    meta: headmeta.Caf
    rescaler: AnnRescaler = None
    v_threshold: int = 0
    bmin: float = 0.1  #: in pixels
    visualizer: CafVisualizer = None
    fill_plan: List[Tuple[int, int, int]] = None

    min_size: ClassVar[int] = 3
    fixed_size: ClassVar[bool] = False
    aspect_ratio: ClassVar[float] = 0.0
    padding: ClassVar[int] = 10

    def __post_init__(self):
        if self.rescaler is None:
            self.rescaler = AnnRescaler(self.meta.stride, self.meta.pose)

        if self.visualizer is None:
            self.visualizer = CafVisualizer(self.meta)

        if self.fill_plan is None:
            self.fill_plan = [
                (caf_i, joint1i - 1, joint2i - 1)
                for caf_i, (joint1i, joint2i) in enumerate(self.meta.skeleton)
            ]

    def __call__(self, image, anns, meta):
        return CafGenerator(self)(image, anns, meta)


class AssociationFiller:
    def __init__(self, config: Caf):
        self.config = config
        self.rescaler = config.rescaler
        self.visualizer = config.visualizer

        self.sparse_skeleton_m1 = (
            np.asarray(config.meta.sparse_skeleton) - 1
            if getattr(config.meta, 'sparse_skeleton', None) is not None
            else None
        )

        if self.config.fixed_size:
            assert self.config.aspect_ratio == 0.0

        LOG.debug('only_in_field_of_view = %s, caf min size = %d',
                  config.meta.only_in_field_of_view,
                  self.config.min_size)

        self.field_shape = None
        self.fields_reg_l = None

    def init_fields(self, bg_mask):
        raise NotImplementedError

    def all_fill_values(self, keypoint_sets, anns):
        """Values in the same order and length as keypoint_sets."""
        raise NotImplementedError

    def fill_field_values(self, field_i, fij, fill_values):
        raise NotImplementedError

    def fields_as_tensor(self, valid_area):
        raise NotImplementedError

    def __call__(self, image, anns, meta):
        width_height_original = image.shape[2:0:-1]

        keypoint_sets = self.rescaler.keypoint_sets(anns)
        bg_mask = self.rescaler.bg_mask(anns, width_height_original,
                                        crowd_margin=(self.config.min_size - 1) / 2)
        self.field_shape = (
            self.config.meta.n_fields,
            bg_mask.shape[0] + 2 * self.config.padding,
            bg_mask.shape[1] + 2 * self.config.padding,
        )
        valid_area = self.rescaler.valid_area(meta)
        LOG.debug('valid area: %s', valid_area)

        self.init_fields(bg_mask)
        self.fields_reg_l = np.full(self.field_shape, np.inf, dtype=np.float32)
        p = self.config.padding
        self.fields_reg_l[:, p:-p, p:-p][:, bg_mask == 0] = 1.0

        fill_values = self.all_fill_values(keypoint_sets, anns)
        self.fill(keypoint_sets, fill_values)
        fields = self.fields_as_tensor(valid_area)

        self.visualizer.processed_image(image)
        self.visualizer.targets(fields, annotation_dicts=anns)

        return fields

    def fill(self, keypoint_sets, fill_values):
        for keypoints, fill_value in zip(keypoint_sets, fill_values):
            self.fill_keypoints(keypoints, fill_value)

    def shortest_sparse(self, joint_i, keypoints):
        shortest = np.inf
        for joint1i, joint2i in self.sparse_skeleton_m1:
            if joint_i not in (joint1i, joint2i):
                continue

            joint1 = keypoints[joint1i]
            joint2 = keypoints[joint2i]
            if joint1[2] <= self.config.v_threshold or joint2[2] <= self.config.v_threshold:
                continue

            d = np.linalg.norm(joint1[:2] - joint2[:2])
            shortest = min(d, shortest)

        return shortest

    def fill_keypoints(self, keypoints, fill_values):
        for field_i, joint1i, joint2i in self.config.fill_plan:
            joint1 = keypoints[joint1i]
            joint2 = keypoints[joint2i]
            if joint1[2] <= self.config.v_threshold or joint2[2] <= self.config.v_threshold:
                continue

            # check if there are shorter connections in the sparse skeleton
            if self.sparse_skeleton_m1 is not None:
                d = (np.linalg.norm(joint1[:2] - joint2[:2])
                     / self.config.meta.dense_to_sparse_radius)
                if self.shortest_sparse(joint1i, keypoints) < d \
                   and self.shortest_sparse(joint2i, keypoints) < d:
                    continue

            # if there is no continuous visual connection, endpoints outside
            # the field of view cannot be inferred
            # LOG.debug('fov check: j1 = %s, j2 = %s', joint1, joint2)
            out_field_of_view_1 = (
                joint1[0] < 0
                or joint1[1] < 0
                or joint1[0] > self.field_shape[2] - 1 - 2 * self.config.padding
                or joint1[1] > self.field_shape[1] - 1 - 2 * self.config.padding
            )
            out_field_of_view_2 = (
                joint2[0] < 0
                or joint2[1] < 0
                or joint2[0] > self.field_shape[2] - 1 - 2 * self.config.padding
                or joint2[1] > self.field_shape[1] - 1 - 2 * self.config.padding
            )
            if out_field_of_view_1 and out_field_of_view_2:
                continue
            if self.config.meta.only_in_field_of_view:
                if out_field_of_view_1 or out_field_of_view_2:
                    continue

            self.fill_association(field_i, joint1, joint2, fill_values)

    def fill_association(self, field_i, joint1, joint2, fill_values):
        # offset between joints
        offset = joint2[:2] - joint1[:2]
        offset_d = np.linalg.norm(offset)

        # dynamically create s
        s = max(self.config.min_size, int(offset_d * self.config.aspect_ratio))

        # meshgrid: coordinate matrix
        xyv = np.stack(np.meshgrid(
            np.linspace(-0.5 * (s - 1), 0.5 * (s - 1), s),
            np.linspace(-0.5 * (s - 1), 0.5 * (s - 1), s),
        ), axis=-1).reshape(-1, 2)

        # set fields
        num = max(2, int(np.ceil(offset_d)))
        fmargin = (s / 2) / (offset_d + np.spacing(1))
        fmargin = np.clip(fmargin, 0.25, 0.4)
        # fmargin = 0.0
        frange = np.linspace(fmargin, 1.0 - fmargin, num=num)
        if self.config.fixed_size:
            frange = [0.5]
        filled_ij = set()
        for f in frange:
            for xyo in xyv:
                fij = np.round(joint1[:2] + f * offset + xyo).astype(np.int) + self.config.padding
                if fij[0] < 0 or fij[0] >= self.field_shape[2] or \
                   fij[1] < 0 or fij[1] >= self.field_shape[1]:
                    continue

                # convert to hashable coordinate and check whether
                # it was processed before
                fij_int = (int(fij[0]), int(fij[1]))
                if fij_int in filled_ij:
                    continue
                filled_ij.add(fij_int)

                # mask
                # perpendicular distance computation:
                # https://en.wikipedia.org/wiki/Distance_from_a_point_to_a_line
                # Coordinate systems for this computation is such that
                # joint1 is at (0, 0).
                fxy = fij - self.config.padding
                f_offset = fxy - joint1[:2]
                sink_l = np.fabs(
                    offset[1] * f_offset[0]
                    - offset[0] * f_offset[1]
                ) / (offset_d + 0.01)
                if sink_l > self.fields_reg_l[field_i, fij[1], fij[0]]:
                    continue
                self.fields_reg_l[field_i, fij[1], fij[0]] = sink_l

                self.fill_field_values(field_i, fij, fill_values)


class CafGenerator(AssociationFiller):
    def __init__(self, config: Caf):
        super().__init__(config)

        self.skeleton_m1 = np.asarray(config.meta.skeleton) - 1

        self.intensities = None
        self.fields_reg1 = None
        self.fields_reg2 = None
        self.fields_bmin1 = None
        self.fields_bmin2 = None
        self.fields_scale1 = None
        self.fields_scale2 = None

    def init_fields(self, bg_mask):
        reg_field_shape = (self.field_shape[0], 2, self.field_shape[1], self.field_shape[2])

        self.intensities = np.zeros(self.field_shape, dtype=np.float32)
        self.fields_reg1 = np.full(reg_field_shape, np.nan, dtype=np.float32)
        self.fields_reg2 = np.full(reg_field_shape, np.nan, dtype=np.float32)
        self.fields_bmin1 = np.full(self.field_shape, np.nan, dtype=np.float32)
        self.fields_bmin2 = np.full(self.field_shape, np.nan, dtype=np.float32)
        self.fields_scale1 = np.full(self.field_shape, np.nan, dtype=np.float32)
        self.fields_scale2 = np.full(self.field_shape, np.nan, dtype=np.float32)

        # bg_mask
        p = self.config.padding
        self.intensities[:, p:-p, p:-p][:, bg_mask == 0] = np.nan

    def all_fill_values(self, keypoint_sets, anns):
        return [(kps, self.rescaler.scale(kps)) for kps in keypoint_sets]

    def fill_field_values(self, field_i, fij, fill_values):
        joint1i, joint2i = self.skeleton_m1[field_i]
        keypoints, scale = fill_values

        # update intensity
        self.intensities[field_i, fij[1], fij[0]] = 1.0

        # update regressions
        fxy = fij - self.config.padding
        self.fields_reg1[field_i, :, fij[1], fij[0]] = keypoints[joint1i][:2] - fxy
        self.fields_reg2[field_i, :, fij[1], fij[0]] = keypoints[joint2i][:2] - fxy

        # update bmin
        bmin = self.config.bmin / self.config.meta.stride
        self.fields_bmin1[field_i, fij[1], fij[0]] = bmin
        self.fields_bmin2[field_i, fij[1], fij[0]] = bmin

        # update scale
        if self.config.meta.sigmas is None:
            scale1, scale2 = scale, scale
        else:
            scale1 = scale * self.config.meta.sigmas[joint1i]
            scale2 = scale * self.config.meta.sigmas[joint2i]
        assert np.isnan(scale1) or 0.0 < scale1 < 100.0
        self.fields_scale1[field_i, fij[1], fij[0]] = scale1
        assert np.isnan(scale2) or 0.0 < scale2 < 100.0
        self.fields_scale2[field_i, fij[1], fij[0]] = scale2

    def fields_as_tensor(self, valid_area):
        p = self.config.padding
        intensities = self.intensities[:, p:-p, p:-p]
        fields_reg1 = self.fields_reg1[:, :, p:-p, p:-p]
        fields_reg2 = self.fields_reg2[:, :, p:-p, p:-p]
        fields_bmin1 = self.fields_bmin1[:, p:-p, p:-p]
        fields_bmin2 = self.fields_bmin2[:, p:-p, p:-p]
        fields_scale1 = self.fields_scale1[:, p:-p, p:-p]
        fields_scale2 = self.fields_scale2[:, p:-p, p:-p]

        mask_valid_area(intensities, valid_area)
        mask_valid_area(fields_reg1[:, 0], valid_area, fill_value=np.nan)
        mask_valid_area(fields_reg1[:, 1], valid_area, fill_value=np.nan)
        mask_valid_area(fields_reg2[:, 0], valid_area, fill_value=np.nan)
        mask_valid_area(fields_reg2[:, 1], valid_area, fill_value=np.nan)
        mask_valid_area(fields_bmin1, valid_area, fill_value=np.nan)
        mask_valid_area(fields_bmin2, valid_area, fill_value=np.nan)
        mask_valid_area(fields_scale1, valid_area, fill_value=np.nan)
        mask_valid_area(fields_scale2, valid_area, fill_value=np.nan)

        return torch.from_numpy(np.concatenate([
            np.expand_dims(intensities, 1),
            fields_reg1,
            fields_reg2,
            np.expand_dims(fields_bmin1, 1),
            np.expand_dims(fields_bmin2, 1),
            np.expand_dims(fields_scale1, 1),
            np.expand_dims(fields_scale2, 1),
        ], axis=1))



import copy
import logging

from ..preprocess import Preprocess

LOG = logging.getLogger(__name__)


class ImageToTracking(Preprocess):
    def __call__(self, image, anns, meta):
        anns0 = copy.deepcopy(anns)
        anns1 = copy.deepcopy(anns)
        meta0 = copy.deepcopy(meta)
        meta1 = copy.deepcopy(meta)

        meta0['group_i'] = 0
        meta1['group_i'] = 1

        for ann_i, (ann0, ann1) in enumerate(zip(anns0, anns1)):
            ann0['track_id'] = ann_i
            ann1['track_id'] = ann_i

        return [image, image], [anns0, anns1], [meta0, meta1]



import os

import numpy as np
try:
    import matplotlib.cm as mplcm
    from matplotlib.animation import FuncAnimation
    from mpl_toolkits.mplot3d import Axes3D
except ImportError:
    pass

import openpifpaf

from .transforms import transform_skeleton

CAR_KEYPOINTS_24 = [
    'front_up_right',       # 1
    'front_up_left',        # 2
    'front_light_right',    # 3
    'front_light_left',     # 4
    'front_low_right',      # 5
    'front_low_left',       # 6
    'central_up_left',      # 7
    'front_wheel_left',     # 8
    'rear_wheel_left',      # 9
    'rear_corner_left',     # 10
    'rear_up_left',         # 11
    'rear_up_right',        # 12
    'rear_light_left',      # 13
    'rear_light_right',     # 14
    'rear_low_left',        # 15
    'rear_low_right',       # 16
    'central_up_right',     # 17
    'rear_corner_right',    # 18
    'rear_wheel_right',     # 19
    'front_wheel_right',    # 20
    'rear_plate_left',      # 21
    'rear_plate_right',     # 22
    'mirror_edge_left',     # 23
    'mirror_edge_right',    # 24
]

SKELETON_ORIG = [
    [49, 46], [49, 8], [49, 57], [8, 0], [8, 11], [57, 0],
    [57, 52], [0, 5], [52, 5], [5, 7],  # frontal
    [7, 20], [11, 23], [20, 23], [23, 25], [34, 32],
    [9, 11], [9, 7], [9, 20], [7, 0], [9, 0], [9, 8],  # L-lat
    [24, 33], [24, 25], [24, 11], [25, 32], [25, 28],
    [33, 32], [33, 46], [32, 29], [28, 29],  # rear
    [65, 64], [65, 25], [65, 28], [65, 20], [64, 29],
    [64, 32], [64, 37], [29, 37], [28, 20],  # new rear
    [34, 37], [34, 46], [37, 50], [50, 52], [46, 48], [48, 37],
    [48, 49], [50, 57], [48, 57], [48, 50]
]


KPS_MAPPING = [49, 8, 57, 0, 52, 5, 11, 7, 20, 23, 24, 33, 25, 32, 28,
               29, 46, 34, 37, 50, 65, 64, 9, 48]

CAR_SKELETON_24 = transform_skeleton(SKELETON_ORIG, KPS_MAPPING)

CAR_SIGMAS_24 = [0.05] * len(KPS_MAPPING)

split, error = divmod(len(CAR_KEYPOINTS_24), 4)
CAR_SCORE_WEIGHTS_24 = [10.0] * split + [3.0] * split + \
    [1.0] * split + [0.1] * split + [0.1] * error
assert len(CAR_SCORE_WEIGHTS_24) == len(CAR_KEYPOINTS_24)

HFLIP_24 = {
    'front_up_right': 'front_up_left',
    'front_light_right': 'front_light_left',
    'front_low_right': 'front_low_left',
    'central_up_left': 'central_up_right',
    'front_wheel_left': 'front_wheel_right',
    'rear_wheel_left': 'rear_wheel_right',
    'rear_corner_left': 'rear_corner_right',
    'rear_up_left': 'rear_up_right',
    'rear_light_left': 'rear_light_right',
    'rear_low_left': 'rear_low_right',
    'front_up_left': 'front_up_right',
    'front_light_left': 'front_light_right',
    'front_low_left': 'front_low_right',
    'central_up_right': 'central_up_left',
    'front_wheel_right': 'front_wheel_left',
    'rear_wheel_right': 'rear_wheel_left',
    'rear_corner_right': 'rear_corner_left',
    'rear_up_right': 'rear_up_left',
    'rear_light_right': 'rear_light_left',
    'rear_low_right': 'rear_low_left',
    'rear_plate_left': 'rear_plate_right',
    'rear_plate_right': 'rear_plate_left',
    'mirror_edge_left': 'mirror_edge_right',
    'mirror_edge_right': 'mirror_edge_left'
}

CAR_CATEGORIES_24 = ['car']

p = 0.25
FRONT = -6.0
BACK = 4.5

# CAR POSE is used for joint rescaling. x = [-3, 3] y = [0,4]
CAR_POSE_24 = np.array([
    [-2.9, 4.0, FRONT * 0.5],  # 'front_up_right',              # 1
    [2.9, 4.0, FRONT * 0.5],   # 'front_up_left',               # 2
    [-2.0, 2.0, FRONT],  # 'front_light_right',           # 3
    [2.0, 2.0, FRONT],  # 'front_light_left',             # 4
    [-2.5, 0.0, FRONT],  # 'front_low_right',             # 5
    [2.5, 0.0, FRONT],  # 'front_low_left',              # 6
    [2.6, 4.2, 0.0],  # 'central_up_left'     # 7
    [3.2, 0.2, FRONT * 0.7],  # 'front_wheel_left',           # 8
    [3.0, 0.3, BACK * 0.7],   # 'rear_wheel_left'      # 9
    [3.1, 2.1, BACK * 0.5],   # 'rear_corner_left',          # 10
    [2.4, 4.3, BACK * 0.35],  # 'rear_up_left',       # 11
    [-2.4, 4.3, BACK * 0.35],  # 'rear_up_right'      # 12
    [2.5, 2.2, BACK],   # 'rear_light_left',             # 13
    [-2.5, 2.2, BACK],  # 'rear_light_right',            # 14
    [2.1, 0.1, BACK],  # 'rear_low_left',            # 15
    [-2.1, 0.1, BACK],  # 'rear_low_right',          # 16
    [-2.6, 4.2, 0.0],  # 'central_up_right'    # 17
    [-3.1, 2.1, BACK * 0.5],  # 'rear_corner_right',         # 18
    [-3.0, 0.3, BACK * 0.7],  # 'rear_wheel_right'       # 19
    [-3.2, 0.2, FRONT * 0.7],  # 'front_wheel_right',          # 20
    [1.0, 1.3, BACK],  # 'rear_plate_left',              # 21
    [-1.0, 1.3, BACK],  # 'rear_plate_right',            # 22
    [2.8, 3, FRONT * 0.35],  # 'mirror_edge_left'          # 23
    [-2.8, 3, FRONT * 0.35],  # 'mirror_edge_right'        # 24
])

CAR_POSE_FRONT_24 = np.array([
    [-2.0, 4.0, 2.0],  # 'front_up_right',         # 1
    [2.0, 4.0, 2.0],   # 'front_up_left',        # 2
    [-1.3, 2.0, 2.0],  # 'front_light_right',    # 3
    [1.3, 2.0, 2.0],  # 'front_light_left',     # 4
    [-2.2, 0.0, 2.0],  # 'front_low_right',       # 5
    [2.2, 0.0, 2.0],  # 'front_low_left',       # 6
    [2.0 - p / 2, 4.0 + p, 1.0],  # 'central_up_left',      # 7
    [2.0 + p, 0.1 - p / 2, 1.0],  # 'front_wheel_left',     # 8
    [2, 0.1, 0.0],  # 'rear_wheel_left',      # 9
    [2.6, 1.7, 0.0],   # 'rear_corner_left',          # 10
    [2.0, 4.1, 0.0],  # 'rear_up_left',         # 11
    [-2.0, 4.0, 0.0],  # 'rear_up_right',        # 12
    [2.1, 1.9, 0.0],   # 'rear_light_left',      # 13
    [-2.1, 1.9, 0.0],  # 'rear_right_right',     # 14
    [2.4, 0.1, 0.0],  # 'rear_low_left',        # 15
    [-2.4, 0.1, 0.0],  # 'rear_low_right',       # 16
    [-2.0 + p / 2, 4.0 + p, 1.0],  # 'central_up_right',     # 17
    [-2.6, 1.75, 0.0],  # 'rear_corner_right',           # 18
    [-2, 0.0, 0.0],  # 'rear_wheel_right',     # 19
    [-2 - p, 0.0 - p / 2, 1.0],  # 'front_wheel_right',     # 20
])

CAR_POSE_REAR_24 = np.array([
    [-2.0, 4.0, 0.0],  # 'front_up_right',         # 1
    [2.0, 4.0, 0.0],   # 'front_up_left',        # 2
    [-1.3, 2.0, 0.0],  # 'front_light_right',    # 3
    [1.3, 2.0, 0.0],  # 'front_light_left',     # 4
    [-2.2, 0.0, 0.0],  # 'front_low_right',       # 5
    [2.2, 0.0, 0.0],  # 'front_low_left',       # 6
    [-2.0 + p, 4.0 + p, 2.0],  # 'central_up_left',      # 7
    [2, 0.0, 0.0],  # 'front_wheel_left',     # 8
    [2, 0.0, 0.0],  # 'rear_wheel_left',      # 9
    [-1.6 - p, 2.2 - p, 2.0],   # 'rear_corner_left',     # 10
    [-2.0, 4.0, 2.0],  # 'rear_up_left',         # 11
    [2.0, 4.0, 2.0],  # 'rear_up_right',        # 12
    [-1.6, 2.2, 2.0],   # 'rear_light_left',      # 13
    [1.6, 2.2, 2.0],  # 'rear_right_right',     # 14
    [-2.4, 0.0, 2.0],  # 'rear_low_left',        # 15
    [2.4, 0.0, 2.0],  # 'rear_low_right',       # 16
    [2.0 - p, 4.0 + p, 2.0],  # 'central_up_right',     # 17
    [1.6 + p, 2.2 - p, 2.0],  # 'rear_corner_right', # 18
    [-2, 0.0, 0.0],  # 'rear_wheel_right',     # 19
    [-2, 0.0, 0.0],  # 'front_wheel_right',     # 20
])

CAR_POSE_LEFT_24 = np.array([
    [-2.0, 4.0, 0.0],  # 'front_up_right',         # 1
    [0 - 5 * p, 4.0 - p / 2, 2.0],   # 'front_up_left',        # 2
    [-1.3, 2.0, 0.0],  # 'front_light_right',    # 3
    [1.3, 2.0, 0.0],  # 'front_light_left',     # 4
    [-2.2, 0.0, 0.0],  # 'front_low_right',       # 5
    [-4 - 3 * p, 0.0, 2.0],   # 'front_low_left',       # 6
    [0, 4.0, 2.0],  # 'central_up_left',      # 7
    [-4, 0.0, 2.0],  # 'front_wheel_left',     # 8
    [4, 0.0, 2.0],  # 'rear_wheel_left',      # 9
    [5, 2, 2.0],  # 'rear_corner_left',     # 10
    [0 + 5 * p, 4.0 - p / 2, 2.0],  # 'rear_up_left',  # 11
    [2.0, 4.0, 0.0],  # 'rear_up_right',        # 12
    [5 + p, 2 + p, 1.0],   # 'rear_light_left',      # 13
    [1.6, 2.2, 0.0],  # 'rear_right_right',     # 14
    [-2.4, 0.0, 0.0],  # 'rear_low_left',        # 15
    [2.4, 0.0, 0.0],  # 'rear_low_right',       # 16
    [2.0, 4.0, 0.0],  # 'central_up_right',     # 17
    [1.6, 2.2, 0.0],  # 'rear_corner_right', # 18
    [-2, 0.0, 0.0],  # 'rear_wheel_right',     # 19
    [-2, 0.0, 0.0],  # 'front_wheel_right',     # 20
])


CAR_POSE_RIGHT_24 = np.array([
    [0 + 5 * p, 4.0 - p / 2, 2.0],  # 'front_up_right',         # 1
    [0, 4.0, 0.0],   # 'front_up_left',        # 2
    [-1.3, 2.0, 0.0],  # 'front_light_right',    # 3
    [1.3, 2.0, 0.0],  # 'front_light_left',     # 4
    [4 + 3 * p, 0.0, 2.0],  # 'front_low_right',       # 5
    [-4 - 3, 0.0, 0.0],   # 'front_low_left',       # 6
    [0, 4.0, 0.0],  # 'central_up_left',      # 7
    [-4, 0.0, 0.0],  # 'front_wheel_left',     # 8
    [4, 0.0, 0.0],  # 'rear_wheel_left',      # 9
    [5, 2, 0.0],  # 'rear_corner_left',     # 10
    [0 + 5, 4.0, 0.0],  # 'rear_up_left',  # 11
    [0 - 5 * p, 4.0 - p / 2, 2.0],  # 'rear_up_right',        # 12
    [5, 2, 0.0],   # 'rear_light_left',      # 13
    [-5 - p, 2.0 + p, 2.0],  # 'rear_light_right',     # 14
    [-2.4, 0.0, 0.0],  # 'rear_low_left',        # 15
    [2.4, 0.0, 0.0],  # 'rear_low_right',       # 16
    [0.0, 4.0, 2.0],  # 'central_up_right',     # 17
    [-5, 2.0, 2.0],  # 'rear_corner_right', # 18
    [-4, 0.0, 2.0],  # 'rear_wheel_right',     # 19
    [4, 0.0, 2.0],  # 'front_wheel_right',     # 20
])

CAR_KEYPOINTS_66 = [
    "top_left_c_left_front_car_light",      # 0
    "bottom_left_c_left_front_car_light",   # 1
    "top_right_c_left_front_car_light",     # 2
    "bottom_right_c_left_front_car_light",  # 3
    "top_right_c_left_front_fog_light",     # 4
    "bottom_right_c_left_front_fog_light",  # 5
    "front_section_left_front_wheel",       # 6
    "center_left_front_wheel",              # 7
    "top_right_c_front_glass",              # 8
    "top_left_c_left_front_door",           # 9
    "bottom_left_c_left_front_door",        # 10
    "top_right_c_left_front_door",          # 11
    "middle_c_left_front_door",             # 12
    "front_c_car_handle_left_front_door",   # 13
    "rear_c_car_handle_left_front_door",    # 14
    "bottom_right_c_left_front_door",       # 15
    "top_right_c_left_rear_door",           # 16
    "front_c_car_handle_left_rear_door",    # 17
    "rear_c_car_handle_left_rear_door",     # 18
    "bottom_right_c_left_rear_door",        # 19
    "center_left_rear_wheel",               # 20
    "rear_section_left_rear_wheel",         # 21
    "top_left_c_left_rear_car_light",       # 22
    "bottom_left_c_left_rear_car_light",    # 23
    "top_left_c_rear_glass",                # 24
    "top_right_c_left_rear_car_light",      # 25
    "bottom_right_c_left_rear_car_light",   # 26
    "bottom_left_c_trunk",                  # 27
    "Left_c_rear_bumper",                   # 28
    "Right_c_rear_bumper",                  # 29
    "bottom_right_c_trunk",                 # 30
    "bottom_left_c_right_rear_car_light",   # 31
    "top_left_c_right_rear_car_light",      # 32
    "top_right_c_rear_glass",               # 33
    "bottom_right_c_right_rear_car_light",  # 34
    "top_right_c_right_rear_car_light",     # 35
    "rear_section_right_rear_wheel",        # 36
    "center_right_rear_wheel",              # 37
    "bottom_left_c_right_rear_car_door",    # 38
    "rear_c_car_handle_right_rear_car_door",    # 39
    "front_c_car_handle_right_rear_car_door",   # 40
    "top_left_c_right_rear_car_door",       # 41
    "bottom_left_c_right_front_car_door",   # 42
    "rear_c_car_handle_right_front_car_door",   # 43
    "front_c_car_handle_right_front_car_door",  # 44
    "middle_c_right_front_car_door",        # 45
    "top_left_c_right_front_car_door",      # 46
    "bottom_right_c_right_front_car_door",  # 47
    "top_right_c_right_front_car_door",     # 48
    "top_left_c_front_glass",               # 49
    "center_right_front_wheel",             # 50
    "front_section_right_front_wheel",      # 51
    "bottom_left_c_right_fog_light",        # 52
    "top_left_c_right_fog_light",           # 53
    "bottom_left_c_right_front_car_light",  # 54
    "top_left_c_right_front_car_light",     # 55
    "bottom_right_c_right_front_car_light",  # 56
    "top_right_c_right_front_car_light",     # 57
    "top_right_c_front_lplate",             # 58
    "top_left_c_front_lplate",              # 59
    "bottom_right_c_front_lplate",           # 60
    "bottom_left_c_front_lplate",          # 61
    "top_left_c_rear_lplate",               # 62
    "top_right_c_rear_lplate",              # 63
    "bottom_right_c_rear_lplate",           # 64
    "bottom_left_c_rear_lplate", ]            # 65


HFLIP_ids = {
    0: 57,
    1: 56,
    2: 55,
    3: 54,
    4: 53,
    5: 52,
    6: 51,
    7: 50,
    8: 49,
    9: 48,
    10: 47,
    11: 46,
    12: 45,
    13: 44,
    14: 43,
    15: 42,
    16: 41,
    17: 40,
    18: 39,
    19: 38,
    20: 37,
    21: 36,
    22: 35,
    23: 34,
    24: 33,
    25: 32,
    26: 31,
    27: 30,
    28: 29,
    59: 58,
    61: 60,
    62: 63,
    65: 64
}

HFLIP_66 = {}
checklist = []
for ind in HFLIP_ids:
    HFLIP_66[CAR_KEYPOINTS_66[ind]] = CAR_KEYPOINTS_66[HFLIP_ids[ind]]
    HFLIP_66[CAR_KEYPOINTS_66[HFLIP_ids[ind]]] = CAR_KEYPOINTS_66[ind]
    checklist.append(ind)
    checklist.append(HFLIP_ids[ind])
assert sorted(checklist) == list(range(len(CAR_KEYPOINTS_66)))
assert len(HFLIP_66) == len(CAR_KEYPOINTS_66)

CAR_CATEGORIES_66 = ['car']

SKELETON_LEFT = [
    [59, 61], [59, 1], [61, 5], [0, 1], [0, 2], [2, 3], [3, 1], [3, 4], [4, 5],  # front
    [5, 6], [6, 7], [4, 7], [2, 9], [9, 8], [8, 11], [7, 10], [6, 10], [9, 10],  # side front part
    [11, 12], [11, 24], [9, 12], [10, 15], [12, 15],
    [9, 13], [13, 14], [14, 12], [14, 15],  # side middle part
    [24, 16], [12, 16], [12, 17], [17, 18], [18, 16],
    [15, 19], [19, 20], [19, 18], [20, 21], [16, 21],  # side back part
    [16, 22], [21, 28], [22, 23], [23, 28], [22, 25], [25, 26],
    [23, 26], [26, 27], [25, 62], [27, 65], [62, 65], [28, 65]]

SKELETON_RIGHT = [[HFLIP_ids[bone[0]], HFLIP_ids[bone[1]]] for bone in SKELETON_LEFT]

SKELETON_CONNECT = [
    [28, 29], [62, 63], [65, 64], [24, 33], [46, 11],
    [48, 9], [59, 58], [60, 61], [0, 57], [49, 8]]

SKELETON_ALL = SKELETON_LEFT + SKELETON_RIGHT + SKELETON_CONNECT

CAR_SKELETON_66 = [(bone[0] + 1, bone[1] + 1) for bone in SKELETON_ALL]  # COCO style skeleton

CAR_SIGMAS_66 = [0.05] * len(CAR_KEYPOINTS_66)

split, error = divmod(len(CAR_KEYPOINTS_66), 4)
CAR_SCORE_WEIGHTS_66 = [10.0] * split + [3.0] * split + \
    [1.0] * split + [0.1] * split + [0.1] * error
assert len(CAR_SCORE_WEIGHTS_66) == len(CAR_KEYPOINTS_66)


# number plate offsets
P_X = 0.3
P_Y_TOP = -0.2
P_Y_BOTTOM = -0.4

# z for front
FRONT_Z = -2.0
FRONT_Z_SIDE = -1.8
FRONT_Z_CORNER = -1.7
FRONT_Z_WHEEL = -1.4
FRONT_Z_DOOR = -1.0

# lights x offset
LIGHT_X_INSIDE = 0.8
X_OUTSIDE = 1.0

# y offsets
TOP_CAR = 0.5
BOTTOM_LINE = -0.75
TOP_LINE = 0.1

# z for the back
BACK_Z_WHEEL = 1.0
BACK_Z = 1.5
BACK_Z_SIDE = 1.3

CAR_POSE_HALF = np.array([
    [-LIGHT_X_INSIDE, 0.0, FRONT_Z],    # 0
    [-LIGHT_X_INSIDE, -0.2, FRONT_Z],  # 1
    [-X_OUTSIDE, 0.0, FRONT_Z_SIDE],  # 2
    [-X_OUTSIDE, -0.2, FRONT_Z_SIDE],  # 3
    [-X_OUTSIDE, P_Y_BOTTOM, FRONT_Z_SIDE],  # 4
    [-X_OUTSIDE, P_Y_BOTTOM - 0.2, FRONT_Z_SIDE],  # 5
    [-X_OUTSIDE, BOTTOM_LINE, FRONT_Z_CORNER],  # 6
    [-X_OUTSIDE, BOTTOM_LINE + 0.1, FRONT_Z_WHEEL],  # 7
    [-X_OUTSIDE + 0.1, TOP_CAR, FRONT_Z_DOOR + 0.5],  # 8
    [-X_OUTSIDE, TOP_LINE, FRONT_Z_DOOR],  # 9
    [-X_OUTSIDE, BOTTOM_LINE, FRONT_Z_DOOR],  # 10
    [-X_OUTSIDE + 0.1, TOP_CAR, 0.1],  # 11
    [-X_OUTSIDE, TOP_LINE, 0.05],  # 12
    [-X_OUTSIDE, 0.0, -0.1],  # 13
    [-X_OUTSIDE, 0.0, 0.0],  # 14
    [-X_OUTSIDE, BOTTOM_LINE, 0.0],  # 15
    [-X_OUTSIDE, TOP_LINE, BACK_Z_WHEEL],  # 16
    [-X_OUTSIDE, 0.0, BACK_Z_WHEEL * 0.8],  # 17
    [-X_OUTSIDE, 0.0, BACK_Z_WHEEL * 0.9],  # 18
    [-X_OUTSIDE, BOTTOM_LINE, BACK_Z_WHEEL * 0.6],  # 19
    [-X_OUTSIDE, BOTTOM_LINE + 0.1, BACK_Z_WHEEL],  # 20
    [-X_OUTSIDE, BOTTOM_LINE, BACK_Z_SIDE - 0.2],  # 21
    [-X_OUTSIDE, 0.0, BACK_Z_SIDE],  # 22
    [-X_OUTSIDE, -0.2, BACK_Z_SIDE],  # 23
    [-X_OUTSIDE + 0.1, TOP_CAR - 0.1, BACK_Z_WHEEL],  # 24
    [-LIGHT_X_INSIDE, 0.0, BACK_Z],  # 25
    [-LIGHT_X_INSIDE, -0.2, BACK_Z],  # 26
    [-LIGHT_X_INSIDE + 0.1, -0.3, BACK_Z],  # 27
    [-X_OUTSIDE + 0.1, BOTTOM_LINE, BACK_Z]] + \
    [[np.nan, np.nan, np.nan]] * 30 + \
    [[-P_X, P_Y_TOP, FRONT_Z]] + \
    [[np.nan, np.nan, np.nan]] + \
    [[-P_X, P_Y_BOTTOM, FRONT_Z],  # 61
     [-P_X, P_Y_TOP, BACK_Z]] + \
    [[np.nan, np.nan, np.nan]] * 2 + \
    [[-P_X, P_Y_BOTTOM, BACK_Z]])  # 65

CAR_POSE_66 = CAR_POSE_HALF
for key in HFLIP_ids:
    CAR_POSE_66[HFLIP_ids[key], :] = CAR_POSE_HALF[key, :]
    CAR_POSE_66[HFLIP_ids[key], 0] = -CAR_POSE_HALF[key, 0]
assert not np.any(CAR_POSE_66 == np.nan)

training_weights_local_centrality = [
    0.890968488270775,
    0.716506138617812,
    1.05674590410869,
    0.764774195768455,
    0.637682585483328,
    0.686680807728366,
    0.955422595797394,
    0.936714585642375,
    1.34823795445326,
    1.38308992581967,
    1.32689945125819,
    1.38838655605483,
    1.18980184904613,
    1.02584355494795,
    0.90969156732068,
    1.24732068576104,
    1.11338768064342,
    0.933815217550391,
    0.852297518872114,
    1.04167641424727,
    1.01668968075247,
    1.34625964088011,
    0.911796331039028,
    0.866206536337413,
    1.55957820407853,
    0.730844382675724,
    0.651138644197359,
    0.758018559633786,
    1.31842501396691,
    1.32186116654782,
    0.744347016851606,
    0.636390683664723,
    0.715244950821949,
    1.63122349407032,
    0.849835699185461,
    0.910488007220499,
    1.44244151650561,
    1.14150437331681,
    1.19808610191343,
    0.960186788642886,
    1.05023623286937,
    1.19761709710598,
    1.3872216313401,
    1.01256700741214,
    1.1167909667759,
    1.27893496336199,
    1.54475684725655,
    1.40343733870633,
    1.45552060866114,
    1.47264222155031,
    0.970060423999993,
    0.944450314768933,
    0.623987071240172,
    0.5745237907704,
    0.66890646050993,
    0.978411632994504,
    0.587396395188292,
    0.76307999741129,
    0.609793563449648,
    0.67983566494545,
    0.685883538168462,
    0.753587600664775,
    0.770335133588157,
    0.764713638033368,
    0.792364155965385,
    0.796435233566833
]


def get_constants(num_kps):
    if num_kps == 24:
        CAR_POSE_24[:, 2] = 2.0
        return [CAR_KEYPOINTS_24, CAR_SKELETON_24, HFLIP_24, CAR_SIGMAS_24,
                CAR_POSE_24, CAR_CATEGORIES_24, CAR_SCORE_WEIGHTS_24]
    if num_kps == 66:
        CAR_POSE_66[:, 2] = 2.0
        return [CAR_KEYPOINTS_66, CAR_SKELETON_66, HFLIP_66, CAR_SIGMAS_66,
                CAR_POSE_66, CAR_CATEGORIES_66, CAR_SCORE_WEIGHTS_66]
    # using no if-elif-else construction due to pylint no-else-return error
    raise Exception("Only poses with 24 or 66 keypoints are available.")


def draw_ann(ann, *, keypoint_painter, filename=None, margin=0.5, aspect=None, **kwargs):
    from openpifpaf import show  # pylint: disable=import-outside-toplevel

    bbox = ann.bbox()
    xlim = bbox[0] - margin, bbox[0] + bbox[2] + margin
    ylim = bbox[1] - margin, bbox[1] + bbox[3] + margin
    if aspect == 'equal':
        fig_w = 5.0
    else:
        fig_w = 5.0 / (ylim[1] - ylim[0]) * (xlim[1] - xlim[0])

    with show.canvas(filename, figsize=(fig_w, 5), nomargin=True, **kwargs) as ax:
        ax.set_axis_off()
        ax.set_xlim(*xlim)
        ax.set_ylim(*ylim)

        if aspect is not None:
            ax.set_aspect(aspect)

        keypoint_painter.annotation(ax, ann)


def draw_skeletons(pose, sigmas, skel, kps, scr_weights):
    from openpifpaf.annotation import Annotation  # pylint: disable=import-outside-toplevel
    from openpifpaf import show  # pylint: disable=import-outside-toplevel

    scale = np.sqrt(
        (np.max(pose[:, 0]) - np.min(pose[:, 0]))
        * (np.max(pose[:, 1]) - np.min(pose[:, 1]))
    )

    show.KeypointPainter.show_joint_scales = True
    keypoint_painter = show.KeypointPainter()
    ann = Annotation(keypoints=kps, skeleton=skel, score_weights=scr_weights)
    ann.set(pose, np.array(sigmas) * scale)
    os.makedirs('docs', exist_ok=True)
    draw_ann(ann, filename='docs/skeleton_car.png', keypoint_painter=keypoint_painter)


def plot3d_red(ax_2D, p3d, skeleton):
    skeleton = [(bone[0] - 1, bone[1] - 1) for bone in skeleton]

    rot_p90_x = np.array([[1, 0, 0], [0, 0, 1], [0, 1, 0]])
    p3d = p3d @ rot_p90_x

    fig = ax_2D.get_figure()
    ax = Axes3D(fig, auto_add_to_figure=False)
    fig.add_axes(ax)
    ax.set_axis_off()
    ax_2D.set_axis_off()

    ax.view_init(azim=-90, elev=20)
    ax.set_xlabel('X')
    ax.set_ylabel('Y')
    ax.set_zlabel('Z')
    max_range = np.array([p3d[:, 0].max() - p3d[:, 0].min(),
                          p3d[:, 1].max() - p3d[:, 1].min(),
                          p3d[:, 2].max() - p3d[:, 2].min()]).max() / 2.0
    mid_x = (p3d[:, 0].max() + p3d[:, 0].min()) * 0.5
    mid_y = (p3d[:, 1].max() + p3d[:, 1].min()) * 0.5
    mid_z = (p3d[:, 2].max() + p3d[:, 2].min()) * 0.5

    ax.set_xlim(mid_x - max_range, mid_x + max_range)
    ax.set_ylim(mid_y - max_range, mid_y + max_range)
    ax.set_zlim(mid_z - max_range, mid_z + max_range)  # pylint: disable=no-member

    for ci, bone in enumerate(skeleton):
        c = mplcm.get_cmap('tab20')((ci % 20 + 0.05) / 20)  # Same coloring as Pifpaf preds
        ax.plot(p3d[bone, 0], p3d[bone, 1], p3d[bone, 2], color=c)

    def animate(i):
        ax.view_init(elev=10., azim=i)
        return fig

    return FuncAnimation(fig, animate, frames=360, interval=100)


def print_associations():
    print("\nAssociations of the car skeleton with 24 keypoints")
    for j1, j2 in CAR_SKELETON_24:
        print(CAR_KEYPOINTS_24[j1 - 1], '-', CAR_KEYPOINTS_24[j2 - 1])
    print("\nAssociations of the car skeleton with 66 keypoints")
    for j1, j2 in CAR_SKELETON_66:
        print(CAR_KEYPOINTS_66[j1 - 1], '-', CAR_KEYPOINTS_66[j2 - 1])


def main():
    print_associations()
# =============================================================================
#     draw_skeletons(CAR_POSE_24, sigmas = CAR_SIGMAS_24, skel = CAR_SKELETON_24,
#                    kps = CAR_KEYPOINTS_24, scr_weights = CAR_SCORE_WEIGHTS_24)
#     draw_skeletons(CAR_POSE_66, sigmas = CAR_SIGMAS_66, skel = CAR_SKELETON_66,
#                    kps = CAR_KEYPOINTS_66, scr_weights = CAR_SCORE_WEIGHTS_66)
# =============================================================================
    with openpifpaf.show.Canvas.blank(nomargin=True) as ax_2D:
        anim_66 = plot3d_red(ax_2D, CAR_POSE_66, CAR_SKELETON_66)
        anim_66.save('openpifpaf/plugins/apollocar3d/docs/CAR_66_Pose.gif', fps=30)
    with openpifpaf.show.Canvas.blank(nomargin=True) as ax_2D:
        anim_24 = plot3d_red(ax_2D, CAR_POSE_24, CAR_SKELETON_24)
        anim_24.save('openpifpaf/plugins/apollocar3d/docs/CAR_24_Pose.gif', fps=30)


if __name__ == '__main__':
    main()



import logging
import torch

LOG = logging.getLogger(__name__)


class Shell(torch.nn.Module):
    def __init__(self, base_net, head_nets, *,
                 process_input=None, process_heads=None):
        super().__init__()

        self.base_net = base_net
        self.head_nets = None
        self.process_input = process_input
        self.process_heads = process_heads

        self.set_head_nets(head_nets)

    @property
    def head_metas(self):
        if self.head_nets is None:
            return None
        return [hn.meta for hn in self.head_nets]

    def set_head_nets(self, head_nets):
        if not isinstance(head_nets, torch.nn.ModuleList):
            head_nets = torch.nn.ModuleList(head_nets)

        for hn_i, hn in enumerate(head_nets):
            hn.meta.head_index = hn_i
            hn.meta.base_stride = self.base_net.stride

        self.head_nets = head_nets

    def forward(self, image_batch, *, head_mask=None):
        if self.process_input is not None:
            image_batch = self.process_input(image_batch)

        x = self.base_net(image_batch)
        if head_mask is not None:
            head_outputs = tuple(hn(x) if m else None for hn, m in zip(self.head_nets, head_mask))
        else:
            head_outputs = tuple(hn(x) for hn in self.head_nets)

        if self.process_heads is not None:
            head_outputs = self.process_heads(head_outputs)

        return head_outputs


class CrossTalk(torch.nn.Module):
    def __init__(self, strength=0.2):
        super().__init__()
        self.strength = strength

    def forward(self, image_batch):
        if self.training and self.strength:
            rolled_images = torch.cat((image_batch[-1:], image_batch[:-1]))
            image_batch += rolled_images * self.cross_talk
        return image_batch


def model_defaults(net_cpu):
    for m in net_cpu.modules():
        if isinstance(m, (torch.nn.BatchNorm1d, torch.nn.BatchNorm2d)):
            # avoid numerical instabilities
            # (only seen sometimes when training with GPU)
            # Variances in pretrained models can be as low as 1e-17.
            # m.running_var.clamp_(min=1e-8)
            # m.eps = 1e-3  # tf default is 0.001
            # m.eps = 1e-5  # pytorch default

            # This epsilon only appears inside a sqrt in the denominator,
            # i.e. the effective epsilon for division is much bigger than the
            # given eps.
            # See equation here:
            # https://pytorch.org/docs/stable/generated/torch.nn.BatchNorm2d.html
            m.eps = max(m.eps, 1e-3)  # mobilenetv3 actually has 1e-3

            # smaller step size for running std and mean update
            m.momentum = 0.01  # tf default is 0.99
            # m.momentum = 0.1  # pytorch default

        elif isinstance(m, (torch.nn.GroupNorm, torch.nn.LayerNorm)):
            m.eps = 1e-4

        elif isinstance(m, (torch.nn.InstanceNorm1d, torch.nn.InstanceNorm2d)):
            m.eps = 1e-4
            m.momentum = 0.01



import copy
import numpy as np

from .base import Base
from .caf import Caf
from .. import headmeta


class Tcaf(Base):
    def __init__(self, meta: headmeta.Tcaf):
        super().__init__(meta.name)
        self.meta = meta
        self.caf_visualizer = Caf(meta)

    @staticmethod
    def merge_anns(ann1, ann2):
        m = copy.deepcopy(ann1)
        m['keypoints'] = np.concatenate((ann1['keypoints'], ann2['keypoints']), axis=0)
        return m

    def targets(self, field, *, annotation_dicts):
        anns1, anns2 = annotation_dicts

        anns1_by_trackid = {ann['track_id']: ann for ann in anns1}
        merged_annotations = [
            self.merge_anns(anns1_by_trackid[ann2['track_id']], ann2)
            for ann2 in anns2
            if (not ann2['iscrowd']
                and ann2['track_id'] in anns1_by_trackid)
        ]

        self.caf_visualizer.targets(field, annotation_dicts=merged_annotations)

    def predicted(self, field):
        self.caf_visualizer.predicted(field)



"""Evaluation on COCO data."""

import argparse
from collections import defaultdict
import glob
import json
import logging
import os
import sys
import time

import PIL
import torch

from . import datasets, decoder, logger, network, show, visualizer, __version__
from .predictor import Predictor

try:
    import thop
except ImportError:
    thop = None

LOG = logging.getLogger(__name__)


def default_output_name(args):
    output = '{}.eval-{}'.format(network.Factory.checkpoint, args.dataset)

    # coco
    if args.coco_eval_orientation_invariant or args.coco_eval_extended_scale:
        output += '-coco'
        if args.coco_eval_orientation_invariant:
            output += 'o'
        if args.coco_eval_extended_scale:
            output += 's'
    if args.coco_eval_long_edge is not None and args.coco_eval_long_edge != 641:
        output += '-cocoedge{}'.format(args.coco_eval_long_edge)

    # dense
    if args.dense_connections:
        output += '-dense'
        if args.dense_connections != 1.0:
            output += '{}'.format(args.dense_connections)

    return output


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def cli():
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.eval',
        usage='%(prog)s [options]',
        description=__doc__,
        formatter_class=CustomFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=__version__))

    datasets.cli(parser)
    decoder.cli(parser)
    logger.cli(parser)
    network.Factory.cli(parser)
    Predictor.cli(parser, skip_batch_size=True, skip_loader_workers=True)
    show.cli(parser)
    visualizer.cli(parser)

    parser.add_argument('--output', default=None,
                        help='output filename without file extension')
    parser.add_argument('--skip-existing', default=False, action='store_true',
                        help='skip if output eval file exists already')
    parser.add_argument('--no-skip-epoch0', dest='skip_epoch0',
                        default=True, action='store_false',
                        help='do not skip eval for epoch 0')
    parser.add_argument('--watch', default=False, const=60, nargs='?', type=int,
                        help=('Watch a directory for new checkpoint files. '
                              'Optionally specify the number of seconds between checks.')
                        )
    parser.add_argument('--disable-cuda', action='store_true',
                        help='disable CUDA')
    parser.add_argument('--write-predictions', default=False, action='store_true',
                        help='write a json and a zip file of the predictions')
    parser.add_argument('--show-final-image', default=False, action='store_true',
                        help='show the final image')
    parser.add_argument('--show-final-ground-truth', default=False, action='store_true',
                        help='show the final image with ground truth annotations')
    parser.add_argument('--n-images', default=None, type=int)
    parser.add_argument('--loader-warmup', default=3.0, type=float)
    args = parser.parse_args()

    logger.configure(args, LOG)

    # add args.device
    args.device = torch.device('cpu')
    args.pin_memory = False
    if not args.disable_cuda and torch.cuda.is_available():
        args.device = torch.device('cuda')
        args.pin_memory = True
    LOG.debug('neural network device: %s', args.device)

    datasets.configure(args)
    decoder.configure(args)
    network.Factory.configure(args)
    Predictor.configure(args)
    show.configure(args)
    visualizer.configure(args)

    return args


def count_ops(model, height=641, width=641):
    if thop is None:
        print('warning: run "pip3 install thop" to count parameters and ops')
        return -1, -1
    device = next(model.parameters()).device
    dummy_input = torch.randn(1, 3, height, width, device=device)
    gmacs, params = thop.profile(model, inputs=(dummy_input, ))
    LOG.info('GMACs = {0:.2f}, million params = {1:.2f}'.format(gmacs / 1e9, params / 1e6))
    return gmacs, params


# pylint: disable=too-many-statements,too-many-branches
def evaluate(args):
    # generate a default output filename
    if args.output is None:
        args.output = default_output_name(args)

    # skip existing?
    if args.skip_epoch0:
        if network.Factory.checkpoint.endswith('.epoch000'):
            print('Not evaluating epoch 0.')
            return
    if args.skip_existing:
        stats_file = args.output + '.stats.json'
        if os.path.exists(stats_file):
            print('Output file {} exists already. Exiting.'.format(stats_file))
            return
        print('{} not found. Processing: {}'.format(stats_file, network.Factory.checkpoint))

    datamodule = datasets.factory(args.dataset)
    predictor = Predictor(head_metas=datamodule.head_metas)

    data_loader = datamodule.eval_loader()
    prediction_loader = predictor.enumerated_dataloader(enumerate(data_loader))
    if args.loader_warmup:
        LOG.info('Data loader warmup (%.1fs) ...', args.loader_warmup)
        time.sleep(args.loader_warmup)
        LOG.info('Done.')

    metrics = datamodule.metrics()
    total_start = time.perf_counter()
    loop_start = time.perf_counter()

    for image_i, (pred, gt_anns, image_meta) in enumerate(prediction_loader):
        LOG.info('image %d / %d, last loop: %.3fs, images per second=%.1f',
                 image_i, len(data_loader), time.perf_counter() - loop_start,
                 image_i / max(1, (time.perf_counter() - total_start)))
        loop_start = time.perf_counter()

        for metric in metrics:
            metric.accumulate(pred, image_meta, ground_truth=gt_anns)

        if args.show_final_image:
            # show ground truth and predictions on original image
            annotation_painter = show.AnnotationPainter()
            with open(image_meta['local_file_path'], 'rb') as f:
                cpu_image = PIL.Image.open(f).convert('RGB')

            with show.image_canvas(cpu_image) as ax:
                if args.show_final_ground_truth:
                    annotation_painter.annotations(ax, gt_anns, color='grey')
                annotation_painter.annotations(ax, pred)

        if args.n_images is not None and image_i >= args.n_images - 1:
            break

    total_time = time.perf_counter() - total_start

    # model stats
    counted_ops = list(count_ops(predictor.model_cpu))
    local_checkpoint = network.local_checkpoint_path(network.Factory.checkpoint)
    file_size = os.path.getsize(local_checkpoint) if local_checkpoint else -1.0

    # write
    additional_data = {
        'args': sys.argv,
        'version': __version__,
        'dataset': args.dataset,
        'total_time': total_time,
        'checkpoint': network.Factory.checkpoint,
        'count_ops': counted_ops,
        'file_size': file_size,
        'n_images': predictor.total_images,
        'decoder_time': predictor.total_decoder_time,
        'nn_time': predictor.total_nn_time,
    }

    metric_stats = defaultdict(list)
    for metric in metrics:
        if args.write_predictions:
            metric.write_predictions(args.output, additional_data=additional_data)

        this_metric_stats = metric.stats()
        assert (len(this_metric_stats.get('text_labels', []))
                == len(this_metric_stats.get('stats', [])))

        for k, v in this_metric_stats.items():
            metric_stats[k] = metric_stats[k] + v

    stats = dict(**metric_stats, **additional_data)

    # write stats file
    with open(args.output + '.stats.json', 'w') as f:
        json.dump(stats, f)

    LOG.info('stats:\n%s', json.dumps(stats, indent=4))
    LOG.info(
        'time per image: decoder = %.0fms, nn = %.0fms, total = %.0fms',
        1000 * stats['decoder_time'] / stats['n_images'],
        1000 * stats['nn_time'] / stats['n_images'],
        1000 * stats['total_time'] / stats['n_images'],
    )


def watch(args):
    assert args.output is None
    pattern = args.checkpoint
    evaluated_pattern = '{}*eval-{}.stats.json'.format(pattern, args.dataset)

    while True:
        # find checkpoints that have not been evaluated
        all_checkpoints = glob.glob(pattern)
        evaluated = glob.glob(evaluated_pattern)
        if args.skip_epoch0:
            all_checkpoints = [c for c in all_checkpoints
                               if not c.endswith('.epoch000')]
        checkpoints = [c for c in all_checkpoints
                       if not any(e.startswith(c) for e in evaluated)]
        LOG.info('%d checkpoints, %d evaluated, %d todo: %s',
                 len(all_checkpoints), len(evaluated), len(checkpoints), checkpoints)

        # evaluate all checkpoints
        for checkpoint in checkpoints:
            # reset
            args.output = None
            network.Factory.checkpoint = checkpoint

            evaluate(args)

        # wait before looking for more work
        time.sleep(args.watch)


def main():
    args = cli()

    if args.watch:
        watch(args)
    else:
        evaluate(args)


if __name__ == '__main__':
    main()



import numpy as np


class TrackAnnotation:
    track_id_counter = 0

    def __init__(self):
        self.frame_pose = []

        TrackAnnotation.track_id_counter += 1
        self.id_ = TrackAnnotation.track_id_counter

    def add(self, frame_number, pose_annotation):
        self.frame_pose.append((frame_number, pose_annotation))
        return self

    def pose(self, frame_number):
        for frame_i, pose in reversed(self.frame_pose):
            if frame_i < frame_number:
                break
            if frame_i > frame_number:
                continue

            return pose

        return None

    def pose_score(self, frame_number):
        pose = self.pose(frame_number)
        if pose is None:
            return 0.0

        kps_order = np.argsort(pose.data[:, 2])[::-1]
        if pose.data[kps_order[1], 2] < 0.05:
            return 0.0

        pose.score_weights[:] = 1.0
        pose.score_weights[1] = 3.0
        pose.score_weights[2] = 5.0
        pose.score_weights[5:] = 0.1
        pose.score_weights[-2:] = 0.0  # ears are not annotated
        pose.score_weights /= np.sum(pose.score_weights)

        return pose.score

    def score(self, frame_number, current_importance=1.0):
        """Reduce current importance to rank tracks that are still processing
        for the current frame."""
        weights = [1.0 for _ in range(12)]
        weights[0] = current_importance
        return (
            sum(w * self.pose_score(frame_number - i) for i, w in enumerate(weights))
            / sum(weights)
        )

    def __len__(self):
        return len(self.frame_pose)



import logging
from typing import List

from .decoder import Decoder

LOG = logging.getLogger(__name__)


class Multi(Decoder):
    def __init__(self, decoders):
        super().__init__()

        self.decoders = decoders

    def __call__(self, all_fields):
        out = []
        for task_i, decoder in enumerate(self.decoders):
            if decoder is None:
                out.append(None)
                continue
            LOG.debug('task %d', task_i)
            out += decoder(all_fields)

        return out

    def reset(self):
        # TODO: remove?
        for dec in self.decoders:
            if not hasattr(dec, 'reset'):
                continue
            dec.reset()

    @classmethod
    def factory(cls, head_metas) -> List['Generator']:
        raise NotImplementedError



import openpifpaf

from .cocodet import CocoDet
from .cocokp import CocoKp
from .dataset import CocoDataset


def register():
    openpifpaf.DATAMODULES['cocodet'] = CocoDet
    openpifpaf.DATAMODULES['cocokp'] = CocoKp

    # human pose estimation
    openpifpaf.CHECKPOINT_URLS['mobilenetv2'] = (
        'http://github.com/vita-epfl/openpifpaf-torchhub/releases/download/'
        'v0.12a5/mobilenetv2-201112-193315-cocokp-1728a9f5.pkl')
    openpifpaf.CHECKPOINT_URLS['mobilenetv3small'] = (
        'http://github.com/openpifpaf/torchhub/releases/download/v0.13/'
        'mobilenetv3small-210822-213409-cocokp-slurm726252-edge513-o10s-803b24ae.pkl')
    openpifpaf.CHECKPOINT_URLS['mobilenetv3large'] = (
        'http://github.com/openpifpaf/torchhub/releases/download/v0.13/'
        'mobilenetv3large-210820-184901-cocokp-slurm725985-edge513-o10s-6c76cbfb.pkl')
    openpifpaf.CHECKPOINT_URLS['resnet18'] = openpifpaf.PRETRAINED_UNAVAILABLE
    openpifpaf.CHECKPOINT_URLS['resnet50'] = (
        'http://github.com/openpifpaf/torchhub/releases/download/v0.13/'
        'resnet50-210830-150728-cocokp-slurm728641-edge513-o10s-ecd30da4.pkl')
    openpifpaf.CHECKPOINT_URLS['resnet101'] = openpifpaf.PRETRAINED_UNAVAILABLE
    openpifpaf.CHECKPOINT_URLS['resnet152'] = openpifpaf.PRETRAINED_UNAVAILABLE
    openpifpaf.CHECKPOINT_URLS['shufflenetv2x1'] = openpifpaf.PRETRAINED_UNAVAILABLE
    openpifpaf.CHECKPOINT_URLS['shufflenetv2x2'] = openpifpaf.PRETRAINED_UNAVAILABLE
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k16'] = (
        'http://github.com/openpifpaf/torchhub/releases/download/v0.13/'
        'shufflenetv2k16-210820-232500-cocokp-slurm726069-edge513-o10s-7189450a.pkl')
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k16-withdense'] = (
        'http://github.com/vita-epfl/openpifpaf-torchhub/releases/download/'
        'v0.12b4/shufflenetv2k16-210221-131426-cocokp-o10s-627d901e.pkl')
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k30'] = (
        'http://github.com/openpifpaf/torchhub/releases/download/v0.13/'
        'shufflenetv2k30-210821-003923-cocokp-slurm726072-edge513-o10s-5fe1c400.pkl')
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k44'] = openpifpaf.PRETRAINED_UNAVAILABLE
    openpifpaf.CHECKPOINT_URLS['swin_s'] = (
        'http://github.com/dmizr/openpifpaf/releases/download/'
        'v0.12.14/swin_s_fpn_lvl_3_lr_5e-5_resumed-d286d41a.pkl')
    openpifpaf.CHECKPOINT_URLS['swin_b'] = (
        'http://github.com/dmizr/openpifpaf/releases/download/'
        'v0.12.14/swin_b_fpn_lvl_3_lr_5e-5_resumed-fa951ce0.pkl')
    openpifpaf.CHECKPOINT_URLS['swin_t_input_upsample'] = (
        'http://github.com/dmizr/openpifpaf/releases/download/'
        'v0.12.14/swin_t_input_upsample_no_fpn_lr_5e-5_resumed-e0681112.pkl')

    # object detection
    openpifpaf.CHECKPOINT_URLS['mobilenetv3small-cocodet'] = (
        'http://github.com/openpifpaf/torchhub/releases/download/v0.13/'
        'mobilenetv3small-210822-215020-cocodet-slurm726253-5f2c894f.pkl')
    openpifpaf.CHECKPOINT_URLS['resnet18-cocodet'] = (
        'http://github.com/openpifpaf/torchhub/releases/download/'
        'v0.12.10/resnet18-210526-031303-cocodet-slurm610002-1faf5801.pkl')



"""Export a checkpoint as a CoreML model."""

import argparse
import logging

import PIL
import torch

import openpifpaf

from .export_onnx import image_size_warning

try:
    import coremltools
except ImportError:
    coremltools = None

LOG = logging.getLogger(__name__)


def apply(model, outfile, *, input_w=129, input_h=97, minimum_deployment_target='iOS14'):
    assert coremltools is not None
    image_size_warning(model.base_net.stride, input_w, input_h)

    # configure: inplace-ops are not supported
    openpifpaf.network.heads.CompositeField3.inplace_ops = False
    openpifpaf.network.heads.CompositeField4.inplace_ops = False

    dummy_input = torch.randn(1, 3, input_h, input_w)
    with torch.no_grad():
        traced_model = torch.jit.trace(model, dummy_input)

    coreml_model = coremltools.convert(
        traced_model,
        inputs=[coremltools.ImageType(name='image', shape=dummy_input.shape,
                                      bias=[-1.0, -1.0, -1.0], scale=1.0 / 127.0)],
        # classifier_config = ct.ClassifierConfig(class_labels)
        minimum_deployment_target=getattr(coremltools.target, minimum_deployment_target),
    )

    # pylint: disable=protected-access
    coremltools.models.utils.rename_feature(
        coreml_model._spec, coreml_model._spec.description.output[0].name, 'cif_head')
    coremltools.models.utils.rename_feature(
        coreml_model._spec, coreml_model._spec.description.output[1].name, 'caf_head')

    # Meta
    coreml_model.input_description['image'] = 'Input image to be classified'
    coreml_model.output_description['cif_head'] = 'Composite Intensity Field'
    coreml_model.output_description['caf_head'] = 'Composite Association Field'
    coreml_model.author = 'Kreiss, Bertoni, Alahi: Composite Fields for Human Pose Estimation'
    coreml_model.license = 'Please see https://github.com/openpifpaf/openpifpaf'
    coreml_model.short_description = 'Composite Fields for Human Pose Estimation'
    coreml_model.version = openpifpaf.__version__

    coreml_model.save(outfile)

    # test predict
    image_input = PIL.Image.new('RGB', (input_w, input_h))
    test_predict = coreml_model.predict({'image': image_input})
    print('!!!!!!!!', test_predict)


def print_preprocessing_spec(out_name):
    spec = coremltools.models.utils.load_spec(out_name)
    print(spec.neuralNetwork.preprocessing)  # pylint: disable=no-member


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def main():
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.export_coreml',
        description=__doc__,
        formatter_class=CustomFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=openpifpaf.__version__))

    openpifpaf.network.Factory.cli(parser)

    parser.add_argument('--outfile', default='openpifpaf-resnet50.mlmodel')
    parser.add_argument('--input-width', type=int, default=129)
    parser.add_argument('--input-height', type=int, default=97)
    parser.add_argument('--minimum-deployment-target', choices=('iOS13', 'iOS14'), default='iOS14')
    args = parser.parse_args()

    openpifpaf.network.Factory.configure(args)

    model, _ = openpifpaf.network.Factory().factory()

    assert args.outfile.endswith('.mlmodel')
    apply(model, args.outfile,
          input_w=args.input_width, input_h=args.input_height,
          minimum_deployment_target=args.minimum_deployment_target)
    print_preprocessing_spec(args.outfile)


if __name__ == '__main__':
    main()



import openpifpaf

from .module import CrowdPose


def register():
    openpifpaf.CHECKPOINT_URLS['resnet50-crowdpose'] = (
        'http://github.com/vita-epfl/openpifpaf-torchhub/releases/download/'
        'v0.12a7/resnet50-201005-100758-crowdpose-d978a89f.pkl'
    )

    openpifpaf.DATAMODULES['crowdpose'] = CrowdPose



import copy
import logging

import torch
import torchvision

from ..preprocess import Preprocess

LOG = logging.getLogger(__name__)


class Pad(Preprocess):
    def __init__(self, target_size, max_shift):
        if isinstance(target_size, int):
            target_size = (target_size, target_size)
        self.target_size = target_size
        self.max_shift = max_shift

    def __call__(self, images, all_anns, metas):
        metas = copy.deepcopy(metas)
        all_anns = copy.deepcopy(all_anns)

        cam_shift = (torch.rand(2) - 0.5) * 2.0 * self.max_shift
        LOG.debug('max shift = %s, this shift = %s', self.max_shift, cam_shift)
        for meta_i, meta in enumerate(metas):
            LOG.debug('valid area before pad: %s, image size = %s',
                      meta['valid_area'], images[meta_i].size)
            images[meta_i], all_anns[meta_i], ltrb = self.center_pad(
                images[meta_i], all_anns[meta_i], cam_shift * meta.get('group_i', 1.0))
            meta['offset'] -= ltrb[:2]
            meta['valid_area'][:2] += ltrb[:2]
            LOG.debug('valid area after pad: %s, image size = %s',
                      meta['valid_area'], images[meta_i].size)

        return images, all_anns, metas

    def center_pad(self, image, anns, cam_shift):
        w, h = image.size

        if self.target_size[0] > w:
            left = (self.target_size[0] - w) / 2.0 + cam_shift[0]
            left = int(torch.clamp(left, 0, self.target_size[0] - w).item())
            right = torch.scalar_tensor(self.target_size[0] - w - left)
            right = int(torch.clamp(right, 0, self.target_size[0] - w).item())
        else:
            left = 0
            right = 0

        if self.target_size[1] > h:
            top = (self.target_size[1] - h) / 2.0 + cam_shift[1]
            top = int(torch.clamp(top, 0, self.target_size[1] - h).item())
            bottom = torch.scalar_tensor(self.target_size[1] - h - top)
            bottom = int(torch.clamp(bottom, 0, self.target_size[1] - h).item())
        else:
            top = 0
            bottom = 0

        ltrb = (left, top, right, bottom)
        LOG.debug('pad with %s', ltrb)

        # pad image
        image = torchvision.transforms.functional.pad(
            image, ltrb, fill=(124, 116, 104))

        # pad annotations
        for ann in anns:
            ann['keypoints'][:, 0] += ltrb[0]
            ann['keypoints'][:, 1] += ltrb[1]
            ann['bbox'][0] += ltrb[0]
            ann['bbox'][1] += ltrb[1]

        return image, anns, ltrb



from .preprocess import Preprocess


class MultiScale(Preprocess):
    def __init__(self, preprocess_list):
        """Create lists of preprocesses.

        Must be the most outer preprocess function.
        Preprocess_list can contain transforms.Compose() functions.
        """
        self.preprocess_list = preprocess_list

    def __call__(self, image, anns, meta):
        image_list, anns_list, meta_list = [], [], []
        for p in self.preprocess_list:
            this_image, this_anns, this_meta = p(image, anns, meta)
            image_list.append(this_image)
            anns_list.append(this_anns)
            meta_list.append(this_meta)

        return image_list, anns_list, meta_list



from .base import Base


def cli(parser):
    group = parser.add_argument_group('visualizer')
    group.add_argument('--debug-indices', default=[], nargs='+',
                       help=('Indices of fields to create debug plots for '
                             'of the form headname:fieldindex, e.g. cif:5. '
                             'Optionally, specify the visualization type, '
                             'e.g. cif:5:hr for the high resolution plot only. '
                             'Use comma separation to specify multiple '
                             'head names, field indices or visualization '
                             'types, e.g. cif:5,6:confidence,hr to visualize '
                             'CIF fields 5 and 6 but only show confidence and '
                             'hr.'))


def configure(args):
    # configure visualizer
    Base.set_all_indices(args.debug_indices)



import logging
from typing import List

import numpy as np
import torch

LOG = logging.getLogger(__name__)


class MultiSamplerProxy:
    def __init__(self, loaders: List[torch.utils.data.DataLoader]):
        self.loaders = loaders

    def __getattribute__(self, name):
        if name == 'set_epoch' and not hasattr(self.loaders[0], 'set_epoch'):
            raise AttributeError

        return object.__getattribute__(self, name)

    def set_epoch(self, value):
        for loader_i, loader in enumerate(self.loaders):
            LOG.info('setting epoch %d for loader %d', value, loader_i)
            loader.sampler.set_epoch(value)


class MultiLoader:
    last_task_index = None
    weights = None

    def __init__(self, loaders: List[torch.utils.data.DataLoader], n_heads: int, *, n_batches=None):
        self.loaders = loaders
        self.n_heads = n_heads
        self.sampler = MultiSamplerProxy(loaders)
        self._weights = self.weights

        if self._weights is None:
            self._weights = [1.0 / len(loaders) for _ in range(len(loaders))]
        elif len(self._weights) == len(loaders) - 1:
            self._weights.append(1.0 - sum(self._weights))
        elif len(self._weights) == len(loaders):
            pass
        else:
            raise Exception('invalid dataset weights: {}'.format(self._weights))
        assert all(w > 0.0 for w in self._weights)
        sum_w = sum(self._weights)
        self._weights = [w / sum_w for w in self._weights]
        LOG.info('dataset weights: %s', self._weights)

        self.n_batches = int(min(len(l) / w for l, w in zip(loaders, self._weights)))
        if n_batches:
            self.n_batches = min(self.n_batches, n_batches)

    def __iter__(self):
        loader_iters = [iter(l) for l in self.loaders]
        n_loaded = [0 for _ in self.loaders]
        while True:
            loader_index = int(np.argmin([n / w for n, w in zip(n_loaded, self._weights)]))
            next_batch = next(loader_iters[loader_index], None)
            if next_batch is None:
                break
            n_loaded[loader_index] += 1
            MultiLoader.last_task_index = loader_index

            # convert targets to multi-targets
            image_batch, target_batch, meta_batch = next_batch
            multi_target_batch = [None for _ in range(self.n_heads)]
            for i, tb in zip(meta_batch[0]['head_indices'], target_batch):
                multi_target_batch[i] = tb

            yield image_batch, multi_target_batch, meta_batch

            if sum(n_loaded) >= self.n_batches:
                break

    def __len__(self):
        return self.n_batches



import openpifpaf

from .nuscenes import NuScenes


def register():
    openpifpaf.DATAMODULES['nuscenes'] = NuScenes
    openpifpaf.CHECKPOINT_URLS['shufflenetv2k16-nuscenes'] = 'http://github.com/DuncanZauss/' \
        'openpifpaf_assets/releases/download/v0.1.0/nuscenes_sk16.pkl.epoch150'



import numpy as np

from ..annotation import Annotation, AnnotationCrowd, AnnotationDet
from .preprocess import Preprocess


class ToAnnotations(Preprocess):
    """Convert inputs to annotation objects."""

    def __init__(self, converters):
        self.converters = converters

    def __call__(self, image, anns, meta):
        anns = [
            ann
            for converter in self.converters
            for ann in converter(anns)
        ]
        return image, anns, meta


class ToKpAnnotations:
    """Input to keypoint annotations."""

    def __init__(self, categories, keypoints_by_category, skeleton_by_category):
        self.keypoints_by_category = keypoints_by_category
        self.skeleton_by_category = skeleton_by_category
        self.categories = categories

    def __call__(self, anns):
        return [
            Annotation(
                self.keypoints_by_category[ann['category_id']],
                self.skeleton_by_category[ann['category_id']],
                categories=self.categories,
            )
            .set(
                ann['keypoints'],
                category_id=ann['category_id'],
                fixed_score='',
                fixed_bbox=ann.get('bbox'),
            )
            for ann in anns
            if not ann['iscrowd'] and np.any(ann['keypoints'][2::3] > 0.0)
        ]


class ToDetAnnotations:
    """Input to detection annotations."""

    def __init__(self, categories):
        self.categories = categories

    def __call__(self, anns):
        return [
            AnnotationDet(categories=self.categories)
            .set(
                ann['category_id'],
                None,
                ann['bbox'],
            )
            for ann in anns
            if not ann['iscrowd'] and np.any(ann['bbox'])
        ]


class ToCrowdAnnotations:
    """Input to crowd annotations."""

    def __init__(self, categories):
        self.categories = categories

    def __call__(self, anns):
        return [
            AnnotationCrowd(categories=self.categories)
            .set(
                ann.get('category_id', 1),
                ann['bbox'],
            )
            for ann in anns
            if ann['iscrowd']
        ]



"""Migrate a model."""

import argparse
import logging

import torch

from . import network, __version__


def main():
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.migrate',
        description=__doc__,
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=__version__))

    parser.add_argument('--debug', default=False, action='store_true')
    parser.add_argument('--output')
    network.Factory.cli(parser)
    args = parser.parse_args()

    network.Factory.configure(args)

    if args.checkpoint is None:
        raise Exception('checkpoint must be provided for old model to migrate from')

    if args.output is None:
        args.output = args.checkpoint + '.out.pkl'

    logging.basicConfig(level=logging.INFO if not args.debug else logging.DEBUG)

    # load old model
    checkpoint = torch.load(args.checkpoint)

    # create a new model
    args.checkpoint = None
    new_model, _ = network.Factory().factory()

    # transfer state from old to new model
    new_model.load_state_dict(checkpoint['model'].state_dict())
    checkpoint['model'] = new_model
    torch.save(checkpoint, args.output)


if __name__ == '__main__':
    main()



import random

import numpy as np
import torch

from ..preprocess import Preprocess


class SingleImage(Preprocess):
    def __init__(self, single_image_op):
        self.single_image_op = single_image_op

    def __call__(self, image_group, anns_group, meta_group):
        out_images = []
        out_anns = []
        out_meta = []

        # force the random generators to change
        _ = torch.randint(0, 10, (1,)).item()
        _ = random.random()
        _ = np.random.rand(1)

        for image, anns, meta in zip(image_group, anns_group, meta_group):
            py_rnd_state = random.getstate()
            np_rnd_state = np.random.get_state()
            with torch.random.fork_rng(devices=[]):
                image, anns, meta = self.single_image_op(image, anns, meta)
                out_images.append(image)
                out_anns.append(anns)
                out_meta.append(meta)
            random.setstate(py_rnd_state)
            np.random.set_state(np_rnd_state)

        return out_images, out_anns, out_meta


class Ungroup(Preprocess):
    """During evaluation, tracking datasets produce image groups of length
    one. Ungroup them so that it looks like any other single-image dataset.
    """
    def __call__(self, image_group, anns_group, meta_group):
        assert len(image_group) == 1
        assert len(anns_group) == 1
        assert len(meta_group) == 1
        return image_group[0], anns_group[0], meta_group[0]



import argparse
import logging

from ... import headmeta
from . import components
from .composite import CompositeLoss, CompositeLossByComponent
from .multi_head import MultiHeadLoss, MultiHeadLossAutoTuneKendall, MultiHeadLossAutoTuneVariance

LOG = logging.getLogger(__name__)

#: headmeta class to Loss class
LOSSES = {
    'a': CompositeLossByComponent,  # TODO hack: added to trigger cli() and configure()
    headmeta.Cif: CompositeLoss,
    headmeta.Caf: CompositeLoss,
    headmeta.CifDet: CompositeLoss,
    headmeta.TSingleImageCif: CompositeLoss,
    headmeta.TSingleImageCaf: CompositeLoss,
    headmeta.Tcaf: CompositeLoss,
}
LOSS_COMPONENTS = {
    components.Bce,
    components.SmoothL1,
    components.Scale,
    components.Laplace,
}


class Factory:
    lambdas = None
    component_lambdas = None
    auto_tune_mtl = False
    auto_tune_mtl_variance = False

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('losses')
        group.add_argument('--lambdas', default=cls.lambdas, type=float, nargs='+',
                           help='prefactor for head losses by head')
        group.add_argument('--component-lambdas',
                           default=cls.component_lambdas, type=float, nargs='+',
                           help='prefactor for head losses by component')
        assert not cls.auto_tune_mtl
        group.add_argument('--auto-tune-mtl', default=False, action='store_true',
                           help=('[experimental] use Kendall\'s prescription for '
                                 'adjusting the multitask weight'))
        assert not cls.auto_tune_mtl_variance
        group.add_argument('--auto-tune-mtl-variance', default=False, action='store_true',
                           help=('[experimental] use Variance prescription for '
                                 'adjusting the multitask weight'))
        assert MultiHeadLoss.task_sparsity_weight == \
            MultiHeadLossAutoTuneKendall.task_sparsity_weight
        assert MultiHeadLoss.task_sparsity_weight == \
            MultiHeadLossAutoTuneVariance.task_sparsity_weight
        group.add_argument('--task-sparsity-weight',
                           default=MultiHeadLoss.task_sparsity_weight, type=float,
                           help='[experimental]')

        for l in set(LOSSES.values()):
            l.cli(parser)
        for lc in LOSS_COMPONENTS:
            lc.cli(parser)

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.lambdas = args.lambdas
        cls.component_lambdas = args.component_lambdas
        cls.auto_tune_mtl = args.auto_tune_mtl
        cls.auto_tune_mtl_variance = args.auto_tune_mtl_variance

        # MultiHeadLoss
        MultiHeadLoss.task_sparsity_weight = args.task_sparsity_weight
        MultiHeadLossAutoTuneKendall.task_sparsity_weight = args.task_sparsity_weight
        MultiHeadLossAutoTuneVariance.task_sparsity_weight = args.task_sparsity_weight

        for l in set(LOSSES.values()):
            l.configure(args)
        for lc in LOSS_COMPONENTS:
            lc.configure(args)

    def factory(self, head_metas):
        sparse_task_parameters = None
        # TODO
        # if MultiHeadLoss.task_sparsity_weight:
        #     sparse_task_parameters = []
        #     for head_net in head_nets:
        #         if getattr(head_net, 'sparse_task_parameters', None) is not None:
        #             sparse_task_parameters += head_net.sparse_task_parameters
        #         elif hasattr(head_net, 'conv'):
        #             sparse_task_parameters.append(head_net.conv.weight)
        #         else:
        #             raise Exception('unknown l1 parameters for given head: {} ({})'
        #                             ''.format(head_net.meta.name, type(head_net)))

        losses = [LOSSES[meta.__class__](meta) for meta in head_metas]
        component_lambdas = self.component_lambdas
        if component_lambdas is None and self.lambdas is not None:
            assert len(self.lambdas) == len(head_metas)
            component_lambdas = [
                head_lambda
                for loss, head_lambda in zip(losses, self.lambdas)
                for _ in loss.field_names
            ]

        if self.auto_tune_mtl:
            loss = MultiHeadLossAutoTuneKendall(
                losses, component_lambdas, sparse_task_parameters=sparse_task_parameters)
        elif self.auto_tune_mtl_variance:
            loss = MultiHeadLossAutoTuneVariance(
                losses, component_lambdas, sparse_task_parameters=sparse_task_parameters)
        else:
            loss = MultiHeadLoss(losses, component_lambdas)

        return loss



import logging
import time

import numpy as np

try:
    import matplotlib
    import matplotlib.animation
    import matplotlib.collections
    import matplotlib.pyplot as plt
    import matplotlib.patches
    from matplotlib.backends.backend_agg import FigureCanvasAgg
except ImportError:
    matplotlib = None
    plt = None

try:
    import pyvirtualcam
except ImportError:
    pyvirtualcam = None

LOG = logging.getLogger(__name__)


class VirtualCamWriter:
    def __init__(self, fps):
        self.fps = fps

        self.cam = None
        self.canvas = None
        self.fig = None

    def setup(self, fig, _, __):
        self.canvas = FigureCanvasAgg(fig)
        self.fig = fig

    def finish(self):
        self.cam.close()

    def grab_frame(self):
        self.canvas.draw()
        f = np.asarray(self.canvas.buffer_rgba())
        LOG.debug('output frame shape: %s', f.shape)

        if self.cam is None:
            assert pyvirtualcam is not None
            self.cam = pyvirtualcam.Camera(f.shape[1], f.shape[0], self.fps)
            LOG.debug('virtual camera: %s', self.cam.device)
        else:
            self.cam.sleep_until_next_frame()
        self.cam.send(f[:, :, :3])


class AnimationFrame:
    """Animations."""

    video_fps = 10  #: frames per second
    video_dpi = 100  #: video dpi
    show = False  #: call matplotlib show()

    def __init__(self, *,
                 fig_width=8.0,
                 fig_init_args=None,
                 video_output=None,
                 second_visual=False):
        self.fig_width = fig_width
        self.fig_init_args = fig_init_args or {}
        self.video_output = video_output
        self.video_writer = None
        if self.video_output == 'virtualcam':
            self.video_writer = VirtualCamWriter(self.video_fps)
        elif self.video_output:
            self.video_writer = matplotlib.animation.writers['ffmpeg'](fps=self.video_fps)

        self.second_visual = second_visual
        if self.second_visual:
            self.fig_width *= 2

        if plt is None:
            LOG.error('matplotlib is not installed')

        self.fig = None
        self.ax = None
        self.ax_second = None
        self._skip_frame = False
        self.last_draw_time = None

        LOG.info('video output = %s', video_output)

    @staticmethod
    def clean_axis(ax):
        xlim = ax.get_xlim()
        ylim = ax.get_ylim()
        ax.cla()
        ax.set_axis_off()
        ax.set_xlim(*xlim)
        ax.set_ylim(*ylim)

    def skip_frame(self):
        self._skip_frame = True

    def iter(self):
        video_writer_is_setup = False
        try:
            while True:
                yield (self.ax, self.ax_second)

                if self._skip_frame:
                    self._skip_frame = False
                    continue

                start = time.perf_counter()

                # Lazy setup of video writer (needs to be after first yield
                # because only that might setup `self.fig`).
                if self.video_writer and not video_writer_is_setup:
                    self.video_writer.setup(self.fig, self.video_output, self.video_dpi)
                    video_writer_is_setup = True

                if self.show:
                    plt.pause(0.01)
                if self.video_writer:
                    self.video_writer.grab_frame()

                if self.ax:
                    self.clean_axis(self.ax)
                if self.ax_second:
                    self.clean_axis(self.ax_second)

                self.last_draw_time = time.perf_counter() - start

        finally:
            if self.video_writer:
                self.video_writer.finish()
            if self.fig:
                plt.close(self.fig)

    def frame_init(self, aspect_ratio=10.0 / 16.0):
        if plt is None:
            return None, None

        if 'figsize' not in self.fig_init_args:
            self.fig_init_args['figsize'] = (
                self.fig_width,
                self.fig_width * aspect_ratio,
            )
            if self.second_visual:
                self.fig_init_args['figsize'] = (
                    self.fig_init_args['figsize'][0],
                    self.fig_init_args['figsize'][1] / 2.0,
                )

        self.fig = plt.figure(**self.fig_init_args)
        if self.second_visual:
            self.ax = plt.Axes(self.fig, [0.0, 0.0, 0.5, 1.0])
            self.ax_second = plt.Axes(self.fig, [0.5, 0.05, 0.45, 0.9])
            self.fig.add_axes(self.ax)
            self.fig.add_axes(self.ax_second)
        else:
            self.ax = plt.Axes(self.fig, [0.0, 0.0, 1.0, 1.0])
            self.ax_second = None
            self.fig.add_axes(self.ax)
        self.ax.set_axis_off()
        if self.ax_second is not None:
            self.ax_second.set_axis_off()

        return self.ax, self.ax_second



import copy
import logging

import numpy as np
import torch

from .. import annotation
from .preprocess import Preprocess

LOG = logging.getLogger(__name__)


class NormalizeAnnotations(Preprocess):
    @classmethod
    def normalize_annotations(cls, anns):
        anns = copy.deepcopy(anns)

        for ann in anns:
            if isinstance(ann, annotation.Base):
                # already converted to an annotation type
                continue

            if 'keypoints' not in ann:
                ann['keypoints'] = []
            if 'iscrowd' not in ann:
                ann['iscrowd'] = False

            ann['keypoints'] = np.asarray(ann['keypoints'], dtype=np.float32).reshape(-1, 3)
            if 'bbox' not in ann:
                ann['bbox'] = cls.bbox_from_keypoints(ann['keypoints'])
            ann['bbox'] = np.asarray(ann['bbox'], dtype=np.float32)
            if 'bbox_original' not in ann:
                ann['bbox_original'] = np.copy(ann['bbox'])
            if 'segmentation' in ann:
                del ann['segmentation']

        return anns

    @staticmethod
    def bbox_from_keypoints(keypoints):
        visible_keypoints = keypoints[keypoints[:, 2] > 0.0]
        if not visible_keypoints.shape[0]:
            return [0, 0, 0, 0]

        x1 = np.min(visible_keypoints[:, 0])
        y1 = np.min(visible_keypoints[:, 1])
        x2 = np.max(visible_keypoints[:, 0])
        y2 = np.max(visible_keypoints[:, 1])
        return [x1, y1, x2 - x2, y2 - y1]

    def __call__(self, image, anns, meta):
        anns = self.normalize_annotations(anns)

        if meta is None:
            meta = {}

        # fill meta with defaults if not already present
        w, h = image.size
        meta_from_image = {
            'offset': np.array((0.0, 0.0)),
            'scale': np.array((1.0, 1.0)),
            'rotation': {'angle': 0.0, 'width': None, 'height': None},
            'valid_area': np.array((0.0, 0.0, w - 1, h - 1)),
            'hflip': False,
            'width_height': np.array((w, h)),
        }
        for k, v in meta_from_image.items():
            if k not in meta:
                meta[k] = v

        return image, anns, meta


class AnnotationJitter(Preprocess):
    def __init__(self, epsilon=0.5):
        self.epsilon = epsilon

    def __call__(self, image, anns, meta):
        meta = copy.deepcopy(meta)
        anns = copy.deepcopy(anns)

        for ann in anns:
            keypoints_xy = ann['keypoints'][:, :2]
            sym_rnd_kp = (torch.rand(*keypoints_xy.shape).numpy() - 0.5) * 2.0
            keypoints_xy += self.epsilon * sym_rnd_kp

            sym_rnd_bbox = (torch.rand((4,)).numpy() - 0.5) * 2.0
            ann['bbox'] += 0.5 * self.epsilon * sym_rnd_bbox

        return image, anns, meta



from .preprocess import Preprocess


class Assert(Preprocess):
    """Inspect (and assert) on current image, anns, meta."""
    def __init__(self, function, message=None):
        self.function = function
        self.message = message

    def __call__(self, *args):
        assert self.function(*args), self.message

        return args



import logging

import numpy as np

from .preprocess import Preprocess

LOG = logging.getLogger(__name__)


class AddCrowdForIncompleteHead(Preprocess):
    @staticmethod
    def __call__(image, anns, meta):
        # detect incomplete annotations and add crowd annotations for those
        new_anns = []
        for ann in anns:
            if ann.get('iscrowd', False):
                continue
            if all(c > 0 for c in ann['keypoints'][0:3, 2]):
                continue

            if 'bbox_head' not in ann:
                LOG.warning('need to add crowd annotation but bbox_head not present')
                continue
            bbox = ann['bbox_head']
            new_anns.append({
                'image_id': ann['image_id'],
                'bbox': bbox.copy(),
                'keypoints': np.array([
                    (bbox[0], bbox[1], 0.0),
                    (bbox[0], bbox[1] + bbox[3], 0.0),
                    (bbox[2], bbox[1] + bbox[3], 0.0),
                    (bbox[2], bbox[1], 0.0),
                ], dtype=np.float32),
                'iscrowd': True,
                'track_id': -1,
            })

        return image, anns + new_anns, meta



import copy
import logging

import numpy as np
import PIL

from .preprocess import Preprocess

LOG = logging.getLogger(__name__)


class Deinterlace(Preprocess):
    def __call__(self, image, anns, meta):
        meta = copy.deepcopy(meta)
        anns = copy.deepcopy(anns)

        # deinterlace image
        w, h = image.size
        image = PIL.Image.fromarray(np.asarray(image)[::2, ::2])
        LOG.debug('before resize = (%f, %f), after = %s', w, h, image.size)

        # rescale keypoints
        for ann in anns:
            ann['keypoints'][:, :2] *= 0.5
            ann['bbox'] *= 0.5

        LOG.debug('meta before: %s', meta)
        meta['offset'] *= 0.5
        meta['scale'] *= 0.5
        meta['valid_area'] *= 0.5
        LOG.debug('meta after: %s', meta)

        return image, anns, meta



"""Plugins for OpenPifPaf.

A plugin is a module that starts with `openpifpaf_`.
The module has to contain a `register()` function.

Follows Flask-style plugin discovery:
https://packaging.python.org/guides/creating-and-discovering-plugins/
"""

import sys
import importlib
import pkgutil

REGISTERED = {}


def register():
    from . import plugins  # pylint: disable=import-outside-toplevel,cyclic-import

    plugin_names = [
        'openpifpaf.plugins.{}'.format(name)
        for finder, name, is_pkg in pkgutil.iter_modules(plugins.__path__)
    ] + [
        name
        for finder, name, is_pkg in pkgutil.iter_modules()
        if name.startswith('openpifpaf_')
    ]

    for name in plugin_names:
        if name not in sys.modules:
            module = importlib.import_module(name)
        else:
            module = sys.modules[name]

        # partially imported modules do not have the register attribute
        if not hasattr(module, 'register'):
            continue

        module.register()
        REGISTERED[name] = module


def versions():
    return {name: getattr(m, '__version__', 'unknown')
            for name, m in REGISTERED.items()
            if not name.startswith('openpifpaf.plugins.')}



from .basenetworks import BaseNetwork
from .running_cache import RunningCache
from ..signal import Signal


class TrackingBase(BaseNetwork):
    cached_items = [0, -1]

    def __init__(self, single_image_backbone):
        super().__init__(
            't' + single_image_backbone.name,
            stride=single_image_backbone.stride,
            out_features=single_image_backbone.out_features,
        )
        self.single_image_backbone = single_image_backbone
        self.running_cache = RunningCache(self.cached_items)

        Signal.subscribe('eval_reset', self.reset)

    def reset(self):
        del self.running_cache
        self.running_cache = RunningCache(self.cached_items)

    def forward(self, *args):
        x = args[0]

        # backbone
        x = self.single_image_backbone(x)

        # feature cache
        if not self.training:
            x = self.running_cache(x)

        return x



"""Export a checkpoint as a TorchScript model."""

import argparse
import logging

import torch

import openpifpaf

from .export_onnx import image_size_warning

LOG = logging.getLogger(__name__)


class DecoderModule(torch.nn.Module):
    def __init__(self, cif_meta, caf_meta):
        super().__init__()

        self.cpp_decoder = torch.classes.openpifpaf_decoder.CifCaf(
            len(cif_meta.keypoints),
            torch.LongTensor(caf_meta.skeleton) - 1,
        )
        self.cif_stride = cif_meta.stride
        self.caf_stride = caf_meta.stride

    def forward(self, cif_field, caf_field):
        return self.cpp_decoder.call(
            cif_field, self.cif_stride,
            caf_field, self.caf_stride,
        )


class EncoderDecoder(torch.nn.Module):
    def __init__(self, traced_encoder, decoder):
        super().__init__()
        self.traced_encoder = traced_encoder
        self.decoder = decoder

    def forward(self, x):
        cif_head_batch, caf_head_batch = self.traced_encoder(x)
        o = [self.decoder(cif_head, caf_head)
             for cif_head, caf_head in zip(cif_head_batch, caf_head_batch)]
        return o


def apply(model, outfile, *, input_w=129, input_h=97):
    image_size_warning(model.base_net.stride, input_w, input_h)

    # configure: inplace-ops are not supported
    openpifpaf.network.heads.CompositeField3.inplace_ops = False
    openpifpaf.network.heads.CompositeField4.inplace_ops = False

    dummy_input = torch.randn(1, 3, input_h, input_w)
    with torch.no_grad():
        traced_encoder = torch.jit.trace(model, dummy_input)
    decoder = DecoderModule(model.head_metas[0], model.head_metas[1])

    encoder_decoder = EncoderDecoder(traced_encoder, decoder)
    encoder_decoder = torch.jit.script(encoder_decoder)

    print(encoder_decoder.graph)
    torch.jit.save(encoder_decoder, outfile)


class CustomFormatter(argparse.ArgumentDefaultsHelpFormatter,
                      argparse.RawDescriptionHelpFormatter):
    pass


def main():
    parser = argparse.ArgumentParser(
        prog='python3 -m openpifpaf.export_torchscript',
        description=__doc__,
        formatter_class=CustomFormatter,
    )
    parser.add_argument('--version', action='version',
                        version='OpenPifPaf {version}'.format(version=openpifpaf.__version__))

    openpifpaf.network.Factory.cli(parser)

    parser.add_argument('--outfile', default='openpifpaf-shufflenetv2k16.torchscript.pt')
    parser.add_argument('--input-width', type=int, default=129)
    parser.add_argument('--input-height', type=int, default=97)
    args = parser.parse_args()

    openpifpaf.network.Factory.configure(args)

    model, _ = openpifpaf.network.Factory().factory()

    assert args.outfile.endswith('.torchscript.pt')
    apply(model, args.outfile, input_w=args.input_width, input_h=args.input_height)


if __name__ == '__main__':
    main()



import argparse
import logging
import time

import numpy as np
import torch

from .. import headmeta, visualizer
from ..annotation import Annotation
from . import utils
from .cifcaf import CifCaf
from .track_annotation import TrackAnnotation
from .track_base import TrackBase

LOG = logging.getLogger(__name__)


class TrackingPose(TrackBase):
    cache_group = [0, -1]
    forward_tracking_pose = True
    track_recovery = False
    single_seed = False

    def __init__(self,
                 cif_meta: headmeta.TSingleImageCif,
                 caf_meta: headmeta.TSingleImageCaf,
                 tcaf_meta: headmeta.Tcaf,
                 *,
                 pose_generator=None):
        super().__init__()
        self.cif_meta = cif_meta
        self.caf_meta = caf_meta
        self.tcaf_meta = tcaf_meta

        # prefer decoders with more keypoints and associations
        self.priority = 1.0
        self.priority += cif_meta.n_fields / 1000.0
        self.priority += caf_meta.n_fields / 1000.0
        self.priority += tcaf_meta.n_fields / 1000.0

        # a similar selector exists for pose similarity
        self.invalid_keypoints = [
            i
            for i, kp in enumerate(cif_meta.keypoints)
            if kp in ('left_ear', 'right_ear')
        ] if cif_meta.dataset == 'posetrack2018' else []

        self.n_keypoints = len(cif_meta.keypoints)
        tracking_keypoints = cif_meta.keypoints * len(self.cache_group)
        tracking_sigmas = cif_meta.sigmas * len(self.cache_group)
        tracking_skeleton = (
            self.caf_meta.skeleton
            + [
                (keypoint_i + 1, keypoint_i + 1 + frame_i * self.n_keypoints)
                for frame_i in range(1, len(self.cache_group))
                for keypoint_i in range(self.n_keypoints)
            ]
        )

        self.tracking_cif_meta = headmeta.Cif(
            'tracking_cif', cif_meta.dataset,
            keypoints=tracking_keypoints,
            sigmas=tracking_sigmas,
            pose=None)
        self.tracking_cif_meta.head_index = 0
        self.tracking_cif_meta.base_stride = cif_meta.base_stride
        self.tracking_cif_meta.upsample_stride = cif_meta.upsample_stride

        self.tracking_caf_meta = headmeta.Caf(
            'tracking_caf', caf_meta.dataset,
            keypoints=tracking_keypoints,
            sigmas=tracking_sigmas,
            skeleton=tracking_skeleton,
            pose=None)
        self.tracking_caf_meta.head_index = 1
        self.tracking_caf_meta.base_stride = caf_meta.base_stride
        self.tracking_caf_meta.upsample_stride = caf_meta.upsample_stride

        self.pose_generator = pose_generator or CifCaf(
            [self.tracking_cif_meta], [self.tracking_caf_meta])
        self.nms_occupancy = utils.Occupancy(2, 4)

        self.vis_multitracking = visualizer.MultiTracking(self.tracking_caf_meta)

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('trackingpose decoder')
        assert not cls.track_recovery
        group.add_argument('--trackingpose-track-recovery', default=False, action='store_true')
        assert not cls.single_seed
        group.add_argument('--trackingpose-single-seed', default=False, action='store_true')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        cls.track_recovery = args.trackingpose_track_recovery
        cls.single_seed = args.trackingpose_single_seed

    @classmethod
    def factory(cls, head_metas):
        if len(head_metas) < 3:
            return []
        return [
            cls(cif_meta, caf_meta, tcaf_meta)
            for cif_meta, caf_meta, tcaf_meta
            in zip(head_metas, head_metas[1:], head_metas[2:])
            if (isinstance(cif_meta, headmeta.TSingleImageCif)
                and isinstance(caf_meta, headmeta.TSingleImageCaf)
                and isinstance(tcaf_meta, headmeta.Tcaf))
        ] + [
            cls(cif_meta, caf_meta, tcaf_meta)
            for cif_meta, caf_meta, tcaf_meta
            in zip(head_metas, head_metas[1:], head_metas[3:])
            if (isinstance(cif_meta, headmeta.TSingleImageCif)
                and isinstance(caf_meta, headmeta.TSingleImageCaf)
                and isinstance(tcaf_meta, headmeta.Tcaf))
        ]

    def soft_nms(self, tracks, frame_number):
        if not tracks:
            return

        # keypoint threshold
        for t in tracks:
            frame_ann = t.pose(self.frame_number)
            if frame_ann is None:
                continue
            kps = frame_ann.data
            kps[kps[:, 2] < utils.nms.Keypoints.get_keypoint_threshold()] = 0.0
            kps[self.invalid_keypoints] = 0.0

        self.nms_occupancy.reset((
            self.n_keypoints,
            int(max(1, max(np.max(t.frame_pose[-1][1].data[:, 1]) for t in tracks) + 1)),
            int(max(1, max(np.max(t.frame_pose[-1][1].data[:, 0]) for t in tracks) + 1)),
        ))

        tracks = sorted(tracks, key=lambda tr: -tr.score(frame_number, current_importance=0.01))
        for track in tracks:
            ann = track.pose(frame_number)
            if ann is None:
                continue

            assert ann.joint_scales is not None
            joint_is = np.flatnonzero(ann.data[:, 2])
            for joint_i in joint_is:
                xyv = ann.data[joint_i]
                if self.nms_occupancy.get(joint_i, xyv[0], xyv[1]):
                    xyv[2] = 0.0
                else:
                    self.nms_occupancy.set(joint_i, xyv[0], xyv[1], ann.joint_scales[joint_i])

        # keypoint threshold
        for t in tracks:
            frame_ann = t.pose(self.frame_number)
            if frame_ann is None:
                continue
            kps = frame_ann.data
            kps[kps[:, 2] < utils.nms.Keypoints.get_keypoint_threshold()] = 0.0

        if self.pose_generator.occupancy_visualizer is not None:
            LOG.debug('Occupied fields after NMS')
            self.pose_generator.occupancy_visualizer.predicted(self.nms_occupancy)

    # pylint: disable=too-many-statements
    def __call__(self, fields, *, initial_annotations=None):
        self.frame_number += 1

        # self.prune_active(frame_number) TODO?

        start = time.perf_counter()

        # initialize tracking poses from self.active tracks
        initial_annotations = []
        for track in self.active:
            tracking_ann = Annotation(
                self.tracking_cif_meta.keypoints,
                self.tracking_caf_meta.skeleton,
            )
            tracking_ann.id_ = track.id_
            for position_i, frame_i in enumerate(self.cache_group[1:], start=1):
                prev_pose = track.pose(self.frame_number + frame_i)
                if prev_pose is not None:
                    tracking_ann.data[
                        self.n_keypoints * position_i:
                        self.n_keypoints * position_i + self.n_keypoints
                    ] = prev_pose.data
                    tracking_ann.joint_scales[
                        self.n_keypoints * position_i:
                        self.n_keypoints * position_i + self.n_keypoints
                    ] = prev_pose.joint_scales

            if self.single_seed:
                inverse_mask = tracking_ann.data[:, 2] < np.amax(tracking_ann.data[:, 2])
                tracking_ann.data[inverse_mask] = 0.0
                tracking_ann.joint_scales[inverse_mask] = 0.0
            tracking_ann.data[tracking_ann.data[:, 2] < 0.05] = 0.0
            if not np.any(tracking_ann.data[:, 2] > 0.0):
                continue
            initial_annotations.append(tracking_ann)
        initial_annotations = list(sorted(
            initial_annotations, key=lambda ann: ann.bbox()[3], reverse=True))
        LOG.debug('initial annotation heights: %s', [ann.bbox()[3] for ann in initial_annotations])

        LOG.debug('using %d initial annotations', len(initial_annotations))

        # use standard pose processor to connect to current frame
        LOG.debug('overwriting CifCaf parameters')
        CifCaf.nms = None
        # CifCaf.keypoint_threshold = 0.001
        tracking_fields = [
            fields[self.cif_meta.head_index],
            torch.cat([
                fields[self.caf_meta.head_index],
                fields[self.tcaf_meta.head_index],
            ], dim=0)
        ]
        tracking_annotations = self.pose_generator(
            tracking_fields, initial_annotations=initial_annotations)

        # extract new pose annotations from tracking pose
        active_by_id = {t.id_: t for t in self.active}
        lost_trackids = {t.id_: t.frame_pose[-1][0] for t in self.active
                         if t.frame_pose[-1][0] < self.frame_number - 1}
        for tracking_ann in tracking_annotations:
            single_frame_ann = Annotation(
                self.cif_meta.keypoints, self.caf_meta.skeleton)
            single_frame_ann.data[:] = tracking_ann.data[:self.n_keypoints]
            single_frame_ann.joint_scales = tracking_ann.joint_scales[:self.n_keypoints]
            single_frame_ann.decoding_order = [
                (jsi, jti, xyv_s, xyv_t)
                for (jsi, jti, xyv_s, xyv_t) in tracking_ann.decoding_order
                if jsi < self.n_keypoints and jti < self.n_keypoints
            ]

            track_id = getattr(tracking_ann, 'id_', -1)
            if track_id == -1:
                new_track = TrackAnnotation().add(self.frame_number, single_frame_ann)
                self.active.append(new_track)
                # assign new track id also to tracking pose for visualization
                tracking_ann.id_ = new_track.id_
                continue
            active_by_id[track_id].add(self.frame_number, single_frame_ann)

        # nms tracks
        self.soft_nms(self.active, self.frame_number)

        # track recovery
        if self.track_recovery:
            removed = set()
            for track in self.active:
                if not lost_trackids:
                    break
                if len(track) > 1:
                    continue
                if track.pose(self.frame_number) is None:
                    continue

                track_id = max(lost_trackids.items(), key=lambda d: d[1])[0]
                del lost_trackids[track_id]
                # tracking_ann.id_ = track_id
                active_by_id[track_id].add(self.frame_number, track.pose(self.frame_number))
                removed.add(track)
                LOG.info('recovered track %d', track_id)
            self.active = [t for t in self.active if t not in removed]

        # tag ignore regions
        # if self.gt_anns:  TODO
        #     self.tag_ignore_region(self.frame_number, self.gt_anns)

        # pruning lost tracks
        self.active = [t for t in self.active if self.track_is_viable(t, self.frame_number)]

        good_track_ids = {t.id_ for t in self.active if self.track_is_good(t, self.frame_number)}
        LOG.info('active tracks = %d, good = %d, track ids = %s',
                 len(self.active), len(good_track_ids),
                 [self.simplified_track_id_map.get(t.id_, t.id_)
                  for t in self.active])

        # visualize good tracking poses with assigned track id
        good_track_annotations = [t for t in tracking_annotations if t.id_ in good_track_ids]
        self.vis_multitracking.predicted(good_track_annotations)
        # if self.track_visualizer or self.track_ann_visualizer:
        #     good_ids = set(t.id_ for t in self.active if self.track_is_good(t, self.frame_number))
        #     good_tracking_anns = [t for t in tracking_annotations
        #                           if getattr(t, 'id_', None) in good_ids]
        #     if self.track_ann_visualizer:
        #         self.track_ann_visualizer.predicted(good_tracking_anns)
        #     if self.track_visualizer:
        #         self.track_visualizer.predicted(
        #             self.frame_number,
        #             [t for t in self.active if t.id_ in good_ids],
        #             self.gt_anns,
        #         )

        LOG.debug('track time: %.3fs', time.perf_counter() - start)
        return self.annotations(self.frame_number)



import argparse

import torch
import numpy as np

import openpifpaf

from openpifpaf.plugins.coco import CocoDataset as Coco
from .wholebody_metric import WholebodyMetric
from .constants import (
    COCO_CATEGORIES,
    WHOLEBODY_KEYPOINTS,
    WHOLEBODY_SKELETON,
    WHOLEBODY_SIGMAS,
    WHOLEBODY_SCORE_WEIGHTS,
    WHOLEBODY_STANDING_POSE,
    HFLIP,
    training_weights_local_centrality
)

try:
    import pycocotools.coco
    # monkey patch for Python 3 compat
    pycocotools.coco.unicode = str
except ImportError:
    pass


class Wholebody(openpifpaf.datasets.DataModule):
    # Coco wholebody has no test dataset

    # cli configurable
    train_annotations = ('data-mscoco/annotations/'
                         'person_keypoints_train2017_wholebody_pifpaf_style.json')
    val_annotations = ('data-mscoco/annotations/'
                       'coco_wholebody_val_v1.0.json')
    train_image_dir = 'data-mscoco/images/train2017/'
    val_image_dir = 'data-mscoco/images/val2017'

    square_edge = 385
    extended_scale = False
    orientation_invariant = 0.0
    blur = 0.0
    augmentation = True
    rescale_images = 1.0
    upsample_stride = 1
    min_kp_anns = 1
    bmin = 1.0

    eval_annotation_filter = True
    eval_long_edge = 641
    eval_orientation_invariant = 0.0
    eval_extended_scale = False
    weights = None

    def __init__(self):
        super().__init__()
        if self.weights is not None:
            caf_weights = []
            for bone in WHOLEBODY_SKELETON:
                caf_weights.append(max(self.weights[bone[0] - 1],
                                       self.weights[bone[1] - 1]))
            w_np = np.array(caf_weights)
            caf_weights = list(w_np / np.sum(w_np) * len(caf_weights))
        else:
            caf_weights = None
        cif = openpifpaf.headmeta.Cif('cif', 'wholebody',
                                      keypoints=WHOLEBODY_KEYPOINTS,
                                      sigmas=WHOLEBODY_SIGMAS,
                                      pose=WHOLEBODY_STANDING_POSE,
                                      draw_skeleton=WHOLEBODY_SKELETON,
                                      score_weights=WHOLEBODY_SCORE_WEIGHTS,
                                      training_weights=self.weights)
        caf = openpifpaf.headmeta.Caf('caf', 'wholebody',
                                      keypoints=WHOLEBODY_KEYPOINTS,
                                      sigmas=WHOLEBODY_SIGMAS,
                                      pose=WHOLEBODY_STANDING_POSE,
                                      skeleton=WHOLEBODY_SKELETON,
                                      training_weights=caf_weights)

        cif.upsample_stride = self.upsample_stride
        caf.upsample_stride = self.upsample_stride
        self.head_metas = [cif, caf]

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('data module wholebody')

        group.add_argument('--wholebody-train-annotations', default=cls.train_annotations,
                           help='train annotations')
        group.add_argument('--wholebody-val-annotations', default=cls.val_annotations,
                           help='val annotations')
        group.add_argument('--wholebody-train-image-dir', default=cls.train_image_dir,
                           help='train image dir')
        group.add_argument('--wholebody-val-image-dir', default=cls.val_image_dir,
                           help='val image dir')

        group.add_argument('--wholebody-square-edge',
                           default=cls.square_edge, type=int,
                           help='square edge of input images')
        assert not cls.extended_scale
        group.add_argument('--wholebody-extended-scale',
                           default=False, action='store_true',
                           help='augment with an extended scale range')
        group.add_argument('--wholebody-orientation-invariant',
                           default=cls.orientation_invariant, type=float,
                           help='augment with random orientations')
        group.add_argument('--wholebody-blur',
                           default=cls.blur, type=float,
                           help='augment with blur')
        assert cls.augmentation
        group.add_argument('--wholebody-no-augmentation',
                           dest='wholebody_augmentation',
                           default=True, action='store_false',
                           help='do not apply data augmentation')
        group.add_argument('--wholebody-rescale-images',
                           default=cls.rescale_images, type=float,
                           help='overall rescale factor for images')
        group.add_argument('--wholebody-upsample',
                           default=cls.upsample_stride, type=int,
                           help='head upsample stride')
        group.add_argument('--wholebody-min-kp-anns',
                           default=cls.min_kp_anns, type=int,
                           help='filter images with fewer keypoint annotations')
        group.add_argument('--wholebody-bmin',
                           default=cls.bmin, type=float,
                           help='bmin')
        group.add_argument('--wholebody-apply-local-centrality-weights',
                           dest='wholebody_apply_local_centrality',
                           default=False, action='store_true',
                           help='Weigh the CIF and CAF head during training.')

        # evaluation
        eval_set_group = group.add_mutually_exclusive_group()
        eval_set_group.add_argument('--wholebody-eval-test2017', default=False,
                                    action='store_true')
        eval_set_group.add_argument('--wholebody-eval-testdev2017', default=False,
                                    action='store_true')

        assert cls.eval_annotation_filter
        group.add_argument('--wholebody-no-eval-annotation-filter',
                           dest='wholebody_eval_annotation_filter',
                           default=True, action='store_false')
        group.add_argument('--wholebody-eval-long-edge', default=cls.eval_long_edge, type=int,
                           help='set to zero to deactivate rescaling')
        assert not cls.eval_extended_scale
        group.add_argument('--wholebody-eval-extended-scale', default=False, action='store_true')
        group.add_argument('--wholebody-eval-orientation-invariant',
                           default=cls.eval_orientation_invariant, type=float)

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.debug = args.debug
        cls.pin_memory = args.pin_memory

        # wholebody specific
        cls.train_annotations = args.wholebody_train_annotations
        cls.val_annotations = args.wholebody_val_annotations
        cls.eval_annotations = cls.val_annotations
        cls.train_image_dir = args.wholebody_train_image_dir
        cls.val_image_dir = args.wholebody_val_image_dir
        cls.eval_image_dir = cls.val_image_dir

        cls.square_edge = args.wholebody_square_edge
        cls.extended_scale = args.wholebody_extended_scale
        cls.orientation_invariant = args.wholebody_orientation_invariant
        cls.blur = args.wholebody_blur
        cls.augmentation = args.wholebody_augmentation
        cls.rescale_images = args.wholebody_rescale_images
        cls.upsample_stride = args.wholebody_upsample
        cls.min_kp_anns = args.wholebody_min_kp_anns
        cls.bmin = args.wholebody_bmin

        # evaluation
        cls.eval_annotation_filter = args.wholebody_eval_annotation_filter
        cls.eval_long_edge = args.wholebody_eval_long_edge
        cls.eval_orientation_invariant = args.wholebody_eval_orientation_invariant
        cls.eval_extended_scale = args.wholebody_eval_extended_scale

        if args.wholebody_apply_local_centrality:
            cls.weights = training_weights_local_centrality
        else:
            cls.weights = None

    def _preprocess(self):
        encoders = (openpifpaf.encoder.Cif(self.head_metas[0], bmin=self.bmin),
                    openpifpaf.encoder.Caf(self.head_metas[1], bmin=self.bmin))

        if not self.augmentation:
            return openpifpaf.transforms.Compose([
                openpifpaf.transforms.NormalizeAnnotations(),
                openpifpaf.transforms.RescaleAbsolute(self.square_edge),
                openpifpaf.transforms.CenterPad(self.square_edge),
                openpifpaf.transforms.EVAL_TRANSFORM,
                openpifpaf.transforms.Encoders(encoders),
            ])

        if self.extended_scale:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.25 * self.rescale_images,
                             2.0 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))
        else:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.4 * self.rescale_images,
                             2.0 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))

        return openpifpaf.transforms.Compose([
            openpifpaf.transforms.NormalizeAnnotations(),
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.HFlip(WHOLEBODY_KEYPOINTS, HFLIP), 0.5),
            rescale_t,
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.Blur(), self.blur),
            openpifpaf.transforms.Crop(self.square_edge, use_area_of_interest=True),
            openpifpaf.transforms.CenterPad(self.square_edge),
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.RotateBy90(), self.orientation_invariant),
            openpifpaf.transforms.TRAIN_TRANSFORM,
            openpifpaf.transforms.Encoders(encoders),
        ])

    def train_loader(self):
        train_data = Coco(
            image_dir=self.train_image_dir,
            ann_file=self.train_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=self.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            train_data, batch_size=self.batch_size, shuffle=not self.debug and self.augmentation,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    def val_loader(self):
        val_data = Coco(
            image_dir=self.val_image_dir,
            ann_file=self.val_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            min_kp_anns=self.min_kp_anns,
            category_ids=[1],
        )
        return torch.utils.data.DataLoader(
            val_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    @classmethod
    def common_eval_preprocess(cls):
        rescale_t = None
        if cls.eval_extended_scale:
            assert cls.eval_long_edge
            rescale_t = [
                openpifpaf.transforms.DeterministicEqualChoice([
                    openpifpaf.transforms.RescaleAbsolute(cls.eval_long_edge),
                    openpifpaf.transforms.RescaleAbsolute((cls.eval_long_edge - 1) // 2 + 1),
                ], salt=1)
            ]
        elif cls.eval_long_edge:
            rescale_t = openpifpaf.transforms.RescaleAbsolute(cls.eval_long_edge)

        if cls.batch_size == 1:
            padding_t = openpifpaf.transforms.CenterPadTight(16)
        else:
            assert cls.eval_long_edge
            padding_t = openpifpaf.transforms.CenterPad(cls.eval_long_edge)

        orientation_t = None
        if cls.eval_orientation_invariant:
            orientation_t = openpifpaf.transforms.DeterministicEqualChoice([
                None,
                openpifpaf.transforms.RotateBy90(fixed_angle=90),
                openpifpaf.transforms.RotateBy90(fixed_angle=180),
                openpifpaf.transforms.RotateBy90(fixed_angle=270),
            ], salt=3)

        return [
            openpifpaf.transforms.NormalizeAnnotations(),
            rescale_t,
            padding_t,
            orientation_t,
        ]

    def _eval_preprocess(self):
        return openpifpaf.transforms.Compose([
            *self.common_eval_preprocess(),
            openpifpaf.transforms.ToAnnotations([
                openpifpaf.transforms.ToKpAnnotations(
                    COCO_CATEGORIES,
                    keypoints_by_category={1: self.head_metas[0].keypoints},
                    skeleton_by_category={1: self.head_metas[1].skeleton},
                ),
                openpifpaf.transforms.ToCrowdAnnotations(COCO_CATEGORIES),
            ]),
            openpifpaf.transforms.EVAL_TRANSFORM,
        ])

    def eval_loader(self):
        eval_data = Coco(
            image_dir=self.eval_image_dir,
            ann_file=self.eval_annotations,
            preprocess=self._eval_preprocess(),
            annotation_filter=self.eval_annotation_filter,
            min_kp_anns=self.min_kp_anns if self.eval_annotation_filter else 0,
            category_ids=[1] if self.eval_annotation_filter else [],
        )
        return torch.utils.data.DataLoader(
            eval_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=False,
            collate_fn=openpifpaf.datasets.collate_images_anns_meta)

    def metrics(self):
        return [
            WholebodyMetric(
                self.eval_annotations,
                max_per_image=20,
                category_ids=[1],
                iou_type='keypoints',
            ),
        ]



"""Head meta objects contain meta information about head networks.

This includes the name, the name of the individual fields, the composition, etc.
"""

from dataclasses import dataclass, field
from typing import Any, ClassVar, List, Tuple

import numpy as np


@dataclass
class Base:
    name: str
    dataset: str

    head_index: int = field(default=None, init=False)
    base_stride: int = field(default=None, init=False)
    upsample_stride: int = field(default=1, init=False)

    @property
    def stride(self) -> int:
        if self.base_stride is None:
            return None
        return self.base_stride // self.upsample_stride

    @property
    def n_fields(self) -> int:
        raise NotImplementedError


@dataclass
class Cif(Base):
    """Head meta data for a Composite Intensity Field (CIF)."""

    keypoints: List[str]
    sigmas: List[float]
    pose: Any = None
    draw_skeleton: List[Tuple[int, int]] = None
    score_weights: List[float] = None

    n_confidences: ClassVar[int] = 1
    n_vectors: ClassVar[int] = 1
    n_scales: ClassVar[int] = 1

    vector_offsets = [True]
    decoder_min_scale = 0.0
    decoder_seed_mask: List[int] = None

    training_weights: List[float] = None

    @property
    def n_fields(self):
        return len(self.keypoints)


@dataclass
class Caf(Base):
    """Head meta data for a Composite Association Field (CAF)."""

    keypoints: List[str]
    sigmas: List[float]
    skeleton: List[Tuple[int, int]]
    pose: Any = None
    sparse_skeleton: List[Tuple[int, int]] = None
    dense_to_sparse_radius: float = 2.0
    only_in_field_of_view: bool = False

    n_confidences: ClassVar[int] = 1
    n_vectors: ClassVar[int] = 2
    n_scales: ClassVar[int] = 2

    vector_offsets = [True, True]
    decoder_min_distance = 0.0
    decoder_max_distance = float('inf')
    decoder_confidence_scales: List[float] = None

    training_weights: List[float] = None

    @property
    def n_fields(self):
        return len(self.skeleton)

    @staticmethod
    def concatenate(metas):
        # TODO: by keypoint name, update skeleton indices if meta.keypoints
        # is not the same for all metas.
        concatenated = Caf(
            name='_'.join(m.name for m in metas),
            dataset=metas[0].dataset,
            keypoints=metas[0].keypoints,
            sigmas=metas[0].sigmas,
            pose=metas[0].pose,
            skeleton=[s for meta in metas for s in meta.skeleton],
            sparse_skeleton=metas[0].sparse_skeleton,
            only_in_field_of_view=metas[0].only_in_field_of_view,
            decoder_confidence_scales=[
                s
                for meta in metas
                for s in (meta.decoder_confidence_scales
                          if meta.decoder_confidence_scales
                          else [1.0 for _ in meta.skeleton])
            ]
        )
        concatenated.head_index = metas[0].head_index
        concatenated.base_stride = metas[0].base_stride
        concatenated.upsample_stride = metas[0].upsample_stride
        return concatenated


@dataclass
class CifDet(Base):
    """Head meta data for a Composite Intensity Field (CIF) for Detection."""

    categories: List[str]

    n_confidences: ClassVar[int] = 1
    n_vectors: ClassVar[int] = 2
    n_scales: ClassVar[int] = 0

    vector_offsets = [True, False]
    decoder_min_scale = 0.0

    training_weights: List[float] = None

    @property
    def n_fields(self):
        return len(self.categories)


@dataclass
class TSingleImageCif(Cif):
    """Single-Image CIF head in tracking models."""


@dataclass
class TSingleImageCaf(Caf):
    """Single-Image CAF head in tracking models."""


@dataclass
class Tcaf(Base):
    """Tracking Composite Association Field."""

    keypoints_single_frame: List[str]
    sigmas_single_frame: List[float]
    pose_single_frame: Any
    draw_skeleton_single_frame: List[Tuple[int, int]] = None
    keypoints: List[str] = None
    sigmas: List[float] = None
    pose: Any = None
    draw_skeleton: List[Tuple[int, int]] = None
    only_in_field_of_view: bool = False

    n_confidences: ClassVar[int] = 1
    n_vectors: ClassVar[int] = 2
    n_scales: ClassVar[int] = 2

    training_weights: List[float] = None

    vector_offsets = [True, True]

    def __post_init__(self):
        if self.keypoints is None:
            self.keypoints = np.concatenate((
                self.keypoints_single_frame,
                self.keypoints_single_frame,
            ), axis=0)
        if self.sigmas is None:
            self.sigmas = np.concatenate((
                self.sigmas_single_frame,
                self.sigmas_single_frame,
            ), axis=0)
        if self.pose is None:
            self.pose = np.concatenate((
                self.pose_single_frame,
                self.pose_single_frame,
            ), axis=0)
        if self.draw_skeleton is None:
            self.draw_skeleton = np.concatenate((
                self.draw_skeleton_single_frame,
                self.draw_skeleton_single_frame,
            ), axis=0)

    @property
    def skeleton(self):
        return [(i + 1, i + 1 + len(self.keypoints_single_frame))
                for i, _ in enumerate(self.keypoints_single_frame)]

    @property
    def n_fields(self):
        return len(self.keypoints_single_frame)



import torch

from .heads import HeadNetwork, CompositeField4


class TBaseSingleImage(HeadNetwork):
    """Filter the feature map so that they can be used by single image loss.

    Training: only apply loss to image 0 of an image pair of image 0 and 1.
    Evaluation with forward tracking pose: only keep image 0.
    Evaluation with full tracking pose: keep all but stack group along feature dim.
    """
    forward_tracking_pose = True
    tracking_pose_length = 2

    def __init__(self, meta, in_features):
        super().__init__(meta, in_features)
        self.head = CompositeField4(meta, in_features)

    def forward(self, *args):
        x = args[0]

        if self.training:
            x = x[::2]
        elif self.forward_tracking_pose:
            x = x[::self.tracking_pose_length]

        x = self.head(x)

        if not self.training and not self.forward_tracking_pose:
            # full tracking pose eval
            # TODO: stack batch dimension in feature dimension and adjust
            # meta information (make it a property to dynamically return
            # a different meta for evaluation)
            raise NotImplementedError

        return x


class Tcaf(HeadNetwork):
    """Filter the feature map so that they can be used by single image loss.

    Training: only apply loss to image 0 of an image pair of image 0 and 1.
    Evaluation with forward tracking pose: only keep image 0.
    Evaluation with full tracking pose: keep all.
    """
    tracking_pose_length = 2
    reduced_features = 512

    _global_feature_reduction = None
    _global_feature_compute = None

    def __init__(self, meta, in_features):
        super().__init__(meta, in_features)

        if self._global_feature_reduction is None:
            self.__class__._global_feature_reduction = torch.nn.Sequential(
                torch.nn.Conv2d(in_features, self.reduced_features,
                                kernel_size=1, bias=True),
                torch.nn.ReLU(inplace=True),
            )
        self.feature_reduction = self._global_feature_reduction

        if self._global_feature_compute is None:
            self.__class__._global_feature_compute = torch.nn.Sequential(
                torch.nn.Conv2d(self.reduced_features * 2, self.reduced_features * 2,
                                kernel_size=1, bias=True),
                torch.nn.ReLU(inplace=True),
            )
        self.feature_compute = self._global_feature_compute

        self.head = CompositeField4(meta, self.reduced_features * 2)

    def forward(self, *args):
        x = args[0]

        # Batches that are not intended for tracking loss might have an
        # odd number of images (or only 1 image).
        # In that case, simply do not execute this head as the result should
        # never be used.
        if len(x) % 2 == 1:
            return None

        x = self.feature_reduction(x)

        group_length = 2 if self.training else self.tracking_pose_length
        primary = x[::group_length]
        others = [x[i::group_length] for i in range(1, group_length)]

        x = torch.stack([torch.cat([primary, o], dim=1) for o in others], dim=1)
        x_shape = x.size()
        x = torch.reshape(x, [x_shape[0] * x_shape[1]] + list(x_shape[2:]))

        x = self.feature_compute(x)

        x = self.head(x)

        if self.tracking_pose_length != 2:
            # TODO need to stack group from batch dim in feature dim and adjust
            # meta info
            raise NotImplementedError

        return x



"""Higher level drawing functions."""

from .base import Base
from .caf import Caf
from .cif import Cif
from .cifdet import CifDet
from .cifhr import CifHr
from .cli import cli, configure
from .multi_tracking import MultiTracking
from .occupancy import Occupancy
from .seeds import Seeds
from .tcaf import Tcaf

__all__ = [
    'Caf',
    'Cif',
    'CifDet',
    'CifHr',
    'MultiTracking',
    'Occupancy',
    'Seeds',
    'Tcaf',
]



import argparse

import torch

import openpifpaf

from .cocokp import CocoKp
from .constants import (
    COCO_CATEGORIES,
    COCO_KEYPOINTS,
    HFLIP,
)
from .dataset import CocoDataset

try:
    import pycocotools.coco
    # monkey patch for Python 3 compat
    pycocotools.coco.unicode = str
except ImportError:
    pass


class CocoDet(openpifpaf.datasets.DataModule):
    # cli configurable
    train_annotations = 'data-mscoco/annotations/instances_train2017.json'
    val_annotations = 'data-mscoco/annotations/instances_val2017.json'
    eval_annotations = val_annotations
    train_image_dir = 'data-mscoco/images/train2017/'
    val_image_dir = 'data-mscoco/images/val2017/'
    eval_image_dir = val_image_dir

    square_edge = 513
    extended_scale = False
    orientation_invariant = 0.0
    blur = 0.0
    augmentation = True
    rescale_images = 1.0
    upsample_stride = 1

    eval_annotation_filter = True

    def __init__(self):
        super().__init__()
        cifdet = openpifpaf.headmeta.CifDet('cifdet', 'cocodet', COCO_CATEGORIES)
        cifdet.upsample_stride = self.upsample_stride
        self.head_metas = [cifdet]

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('data module CocoDet')

        group.add_argument('--cocodet-train-annotations',
                           default=cls.train_annotations)
        group.add_argument('--cocodet-val-annotations',
                           default=cls.val_annotations)
        group.add_argument('--cocodet-train-image-dir',
                           default=cls.train_image_dir)
        group.add_argument('--cocodet-val-image-dir',
                           default=cls.val_image_dir)

        group.add_argument('--cocodet-square-edge',
                           default=cls.square_edge, type=int,
                           help='square edge of input images')
        assert not cls.extended_scale
        group.add_argument('--cocodet-extended-scale',
                           default=False, action='store_true',
                           help='augment with an extended scale range')
        group.add_argument('--cocodet-orientation-invariant',
                           default=cls.orientation_invariant, type=float,
                           help='augment with random orientations')
        group.add_argument('--cocodet-blur',
                           default=cls.blur, type=float,
                           help='augment with blur')
        assert cls.augmentation
        group.add_argument('--cocodet-no-augmentation',
                           dest='cocodet_augmentation',
                           default=True, action='store_false',
                           help='do not apply data augmentation')
        group.add_argument('--cocodet-rescale-images',
                           default=cls.rescale_images, type=float,
                           help='overall rescale factor for images')

        group.add_argument('--cocodet-upsample',
                           default=cls.upsample_stride, type=int,
                           help='head upsample stride')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.debug = args.debug
        cls.pin_memory = args.pin_memory

        # cocodet specific
        cls.train_annotations = args.cocodet_train_annotations
        cls.val_annotations = args.cocodet_val_annotations
        cls.train_image_dir = args.cocodet_train_image_dir
        cls.val_image_dir = args.cocodet_val_image_dir

        cls.square_edge = args.cocodet_square_edge
        cls.extended_scale = args.cocodet_extended_scale
        cls.orientation_invariant = args.cocodet_orientation_invariant
        cls.blur = args.cocodet_blur
        cls.augmentation = args.cocodet_augmentation
        cls.rescale_images = args.cocodet_rescale_images
        cls.upsample_stride = args.cocodet_upsample

        cls.eval_annotation_filter = args.coco_eval_annotation_filter

    def _preprocess(self):
        enc = openpifpaf.encoder.CifDet(self.head_metas[0])

        if not self.augmentation:
            return openpifpaf.transforms.Compose([
                openpifpaf.transforms.NormalizeAnnotations(),
                openpifpaf.transforms.RescaleAbsolute(self.square_edge),
                openpifpaf.transforms.CenterPad(self.square_edge),
                openpifpaf.transforms.EVAL_TRANSFORM,
                openpifpaf.transforms.Encoders([enc]),
            ])

        if self.extended_scale:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.5 * self.rescale_images,
                             2.0 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))
        else:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.7 * self.rescale_images,
                             1.5 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))

        return openpifpaf.transforms.Compose([
            openpifpaf.transforms.NormalizeAnnotations(),
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.HFlip(COCO_KEYPOINTS, HFLIP), 0.5),
            rescale_t,
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.Blur(), self.blur),
            openpifpaf.transforms.RandomChoice(
                [openpifpaf.transforms.RotateBy90(),
                 openpifpaf.transforms.RotateUniform(10.0)],
                [self.orientation_invariant, 0.2],
            ),
            openpifpaf.transforms.Crop(self.square_edge, use_area_of_interest=True),
            openpifpaf.transforms.CenterPad(self.square_edge),
            openpifpaf.transforms.MinSize(min_side=4.0),
            openpifpaf.transforms.UnclippedArea(threshold=0.75),
            openpifpaf.transforms.TRAIN_TRANSFORM,
            openpifpaf.transforms.Encoders([enc]),
        ])

    def train_loader(self):
        train_data = CocoDataset(
            image_dir=self.train_image_dir,
            ann_file=self.train_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            category_ids=[],
        )
        return torch.utils.data.DataLoader(
            train_data, batch_size=self.batch_size, shuffle=not self.debug and self.augmentation,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    def val_loader(self):
        val_data = CocoDataset(
            image_dir=self.val_image_dir,
            ann_file=self.val_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            category_ids=[],
        )
        return torch.utils.data.DataLoader(
            val_data, batch_size=self.batch_size, shuffle=not self.debug and self.augmentation,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    @staticmethod
    def _eval_preprocess():
        return openpifpaf.transforms.Compose([
            *CocoKp.common_eval_preprocess(),
            openpifpaf.transforms.ToAnnotations([
                openpifpaf.transforms.ToDetAnnotations(COCO_CATEGORIES),
                openpifpaf.transforms.ToCrowdAnnotations(COCO_CATEGORIES),
            ]),
            openpifpaf.transforms.EVAL_TRANSFORM,
        ])

    def eval_loader(self):
        eval_data = CocoDataset(
            image_dir=self.eval_image_dir,
            ann_file=self.eval_annotations,
            preprocess=self._eval_preprocess(),
            annotation_filter=self.eval_annotation_filter,
            category_ids=[],
        )
        return torch.utils.data.DataLoader(
            eval_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=False,
            collate_fn=openpifpaf.datasets.collate_images_anns_meta)

    def metrics(self):
        return [openpifpaf.metric.Coco(
            pycocotools.coco.COCO(self.eval_annotations),
            max_per_image=100,
            category_ids=[],
            iou_type='bbox',
        )]



import argparse

import torch

import openpifpaf

from openpifpaf.plugins.coco.cocokp import CocoKp

from openpifpaf.plugins.coco.dataset import CocoDataset

try:
    import pycocotools.coco
    # monkey patch for Python 3 compat
    pycocotools.coco.unicode = str
except ImportError:
    pass

NUSCENES_CATEGORIES = ('truck', 'trailer', 'bus', 'construction_vehicle',
                       'bicycle', 'motorcycle', 'pedestrian', 'traffic_cone',
                       'barrier', 'car')


class NuScenes(openpifpaf.datasets.DataModule):
    # cli configurable
    # MS COCO style annotations created with:
    # https://github.com/open-mmlab/mmdetection3d/blob/master/tools/
    # data_converter/nuimage_converter.py
    train_annotations = '../../../NuScenes/mscoco_style_annotations/nuimages_v1.0-train.json'
    val_annotations = '../../../NuScenes/mscoco_style_annotations/nuimages_v1.0-val.json'
    eval_annotations = '../../../NuScenes/mscoco_style_annotations/nuimages_v1.0-test.json'
    train_image_dir = '../../../NuScenes/nuimages-v1.0-all-samples'
    val_image_dir = train_image_dir    # All images in the the same dir --> annotations
    eval_image_dir = train_image_dir   # are in such a way that the correct file is chosen

    square_edge = 513
    extended_scale = False
    orientation_invariant = 0.0
    blur = 0.0
    augmentation = True
    rescale_images = 1.0
    upsample_stride = 1

    eval_annotation_filter = True

    def __init__(self):
        super().__init__()
        cifdet = openpifpaf.headmeta.CifDet('cifdet', 'nuscenes', NUSCENES_CATEGORIES)
        cifdet.upsample_stride = self.upsample_stride
        self.head_metas = [cifdet]

    @classmethod
    def cli(cls, parser: argparse.ArgumentParser):
        group = parser.add_argument_group('data module NuScenes')

        group.add_argument('--nuscenes-train-annotations',
                           default=cls.train_annotations)
        group.add_argument('--nuscenes-val-annotations',
                           default=cls.val_annotations)
        group.add_argument('--nuscenes-train-image-dir',
                           default=cls.train_image_dir)
        group.add_argument('--nuscenes-val-image-dir',
                           default=cls.val_image_dir)

        group.add_argument('--nuscenes-square-edge',
                           default=cls.square_edge, type=int,
                           help='square edge of input images')
        assert not cls.extended_scale
        group.add_argument('--nuscenes-extended-scale',
                           default=False, action='store_true',
                           help='augment with an extended scale range')
        group.add_argument('--nuscenes-orientation-invariant',
                           default=cls.orientation_invariant, type=float,
                           help='augment with random orientations')
        group.add_argument('--nuscenes-blur',
                           default=cls.blur, type=float,
                           help='augment with blur')
        assert cls.augmentation
        group.add_argument('--nuscenes-no-augmentation',
                           dest='nuscenes_augmentation',
                           default=True, action='store_false',
                           help='do not apply data augmentation')
        group.add_argument('--nuscenes-rescale-images',
                           default=cls.rescale_images, type=float,
                           help='overall rescale factor for images')

        group.add_argument('--nuscenes-upsample',
                           default=cls.upsample_stride, type=int,
                           help='head upsample stride')

    @classmethod
    def configure(cls, args: argparse.Namespace):
        # extract global information
        cls.debug = args.debug
        cls.pin_memory = args.pin_memory

        # nuscenes specific
        cls.train_annotations = args.nuscenes_train_annotations
        cls.val_annotations = args.nuscenes_val_annotations
        cls.train_image_dir = args.nuscenes_train_image_dir
        cls.val_image_dir = args.nuscenes_val_image_dir

        cls.square_edge = args.nuscenes_square_edge
        cls.extended_scale = args.nuscenes_extended_scale
        cls.orientation_invariant = args.nuscenes_orientation_invariant
        cls.blur = args.nuscenes_blur
        cls.augmentation = args.nuscenes_augmentation
        cls.rescale_images = args.nuscenes_rescale_images
        cls.upsample_stride = args.nuscenes_upsample

    def _preprocess(self):
        enc = openpifpaf.encoder.CifDet(self.head_metas[0])

        if not self.augmentation:
            return openpifpaf.transforms.Compose([
                openpifpaf.transforms.NormalizeAnnotations(),
                openpifpaf.transforms.RescaleAbsolute(self.square_edge),
                openpifpaf.transforms.CenterPad(self.square_edge),
                openpifpaf.transforms.EVAL_TRANSFORM,
                openpifpaf.transforms.Encoders([enc]),
            ])

        if self.extended_scale:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.5 * self.rescale_images,
                             2.0 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))
        else:
            rescale_t = openpifpaf.transforms.RescaleRelative(
                scale_range=(0.7 * self.rescale_images,
                             1.5 * self.rescale_images),
                power_law=True, stretch_range=(0.75, 1.33))

        return openpifpaf.transforms.Compose([
            openpifpaf.transforms.NormalizeAnnotations(),
            rescale_t,
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.Blur(), self.blur),
            openpifpaf.transforms.Crop(self.square_edge, use_area_of_interest=True),
            openpifpaf.transforms.CenterPad(self.square_edge),
            openpifpaf.transforms.RandomApply(
                openpifpaf.transforms.RotateBy90(), self.orientation_invariant),
            openpifpaf.transforms.MinSize(min_side=4.0),
            openpifpaf.transforms.UnclippedArea(threshold=0.75),
            openpifpaf.transforms.TRAIN_TRANSFORM,
            openpifpaf.transforms.Encoders([enc]),
        ])

    def train_loader(self):
        train_data = CocoDataset(
            image_dir=self.train_image_dir,
            ann_file=self.train_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            category_ids=[],
        )
        return torch.utils.data.DataLoader(
            train_data, batch_size=self.batch_size, shuffle=not self.debug and self.augmentation,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    def val_loader(self):
        val_data = CocoDataset(
            image_dir=self.val_image_dir,
            ann_file=self.val_annotations,
            preprocess=self._preprocess(),
            annotation_filter=True,
            category_ids=[],
        )
        return torch.utils.data.DataLoader(
            val_data, batch_size=self.batch_size, shuffle=not self.debug and self.augmentation,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=True,
            collate_fn=openpifpaf.datasets.collate_images_targets_meta)

    @staticmethod
    def _eval_preprocess():
        return openpifpaf.transforms.Compose([
            *CocoKp.common_eval_preprocess(),
            openpifpaf.transforms.ToAnnotations([
                openpifpaf.transforms.ToDetAnnotations(NUSCENES_CATEGORIES),
                openpifpaf.transforms.ToCrowdAnnotations(NUSCENES_CATEGORIES),
            ]),
            openpifpaf.transforms.EVAL_TRANSFORM,
        ])

    def eval_loader(self):
        eval_data = CocoDataset(
            image_dir=self.val_image_dir,
            ann_file=self.val_annotations,
            preprocess=self._eval_preprocess(),
            annotation_filter=self.eval_annotation_filter,
            category_ids=[],
        )
        return torch.utils.data.DataLoader(
            eval_data, batch_size=self.batch_size, shuffle=False,
            pin_memory=self.pin_memory, num_workers=self.loader_workers, drop_last=False,
            collate_fn=openpifpaf.datasets.collate_images_anns_meta)

    def metrics(self):
        return [openpifpaf.metric.Coco(
            pycocotools.coco.COCO(self.val_annotations),
            max_per_image=100,
            category_ids=[],
            iou_type='bbox',
        )]
